#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents utility functions.

Functions:
    is_boolean(value: Any) -> bool
        Checks if value type is boolean.
"""
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import re
from ads.common.card_identifier import card_identify
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.feature_engineering.dataset.zip_code_data import zip_code_dict
from functools import lru_cache
from typing import Any


class SchemeNeutral(str):
    BACKGROUND_LIGHT = "#F5F4F2"
    BACKGROUND_DARK = "#E4E1DD"
    AREA_LIGHT = "#BCB6B1"
    AREA_DARK = "#9E9892"
    LINE_LIGHT = "#665F5B"
    LINE_DARK = "#47423E"


class SchemeTeal(str):
    BACKGROUND_LIGHT = "#F0f6f5"
    BACKGROUND_DARK = "#D6E5E5"
    AREA_LIGHT = "#9ABFBF"
    AREA_DARK = "#76A2A0"
    LINE_LIGHT = "#3E686C"
    LINE_DARK = "#2B484B"


def _tag_to_snake(name: str) -> str:
    """Conversts string to snake representation.
    1. Converts the string to the lower case.
    2. Converts all spaces to underscore.

    Parameters
    ----------
    name: string
        The name to convert.

    Returns
    -------
    str: The name converted to the snake representation.
    """
    _name = name.strip().lower()
    if len(_name.split()) > 1:
        _name = "_".join(_name.split())
    return _name


def _add_missing(x, df):
    """
    Adds count of missing values.
    """
    n_missing = pd.isnull(x.replace(r"", np.NaN)).sum()
    if n_missing > 0:
        df.loc["missing"] = n_missing
    return df


def _count_unique_missing(x):
    """
    Returns the total count, unique count and count of missing values of a series.
    """
    df_stat = pd.Series(
        {"count": len(x), "unique": len(x.replace(r"", np.NaN).dropna().unique())},
        name=x.name,
    ).to_frame()
    return _add_missing(x, df_stat)


def is_boolean(value: Any) -> bool:
    """Checks if value type is boolean.

    Parameters
    ----------
    value: Any
        The value to check.

    Returns
    -------
    bool: True if value is boolean, False otherwise.
    """
    bool_values = ("yes", "y", "true", "t", "1", "no", "n", "false", "f", "0", "")
    return isinstance(value, bool) or str(value).lower() in bool_values


def assign_issuer(cardnumber):
    if pd.isnull(cardnumber):
        return "missing"
    else:
        return card_identify().identify_issue_network(cardnumber)


def random_color_func(
    z,
    word=None,
    font_size=None,
    position=None,
    orientation=None,
    font_path=None,
    random_state=None,
):
    """
    Returns random color function use for color_func in creating WordCloud
    """
    h = 179
    s = 23
    l = int(100.0 * float(random_state.randint(60, 120)) / 255.0)
    return "hsl({}, {}%, {}%)".format(h, s, l)


def _is_float(s: str):
    """
    Checks if given string can convert to float
    """
    return re.match(r"^-?\d+(?:\.\d+)$", s) != None


def _str_lat_long_to_point(s):
    """
    Converts input data into formated geometry point
    Return formated geometry point string or np.NaN if input string is not valid
    """
    if isinstance(s, str):
        coords = s.split(",")
        if len(coords) == 2:
            lat, long = coords[0].lstrip(), coords[1].lstrip()
            if lat.startswith("("):
                lat = lat[1:]
            if long.endswith(")"):
                long = long[:-1]
            if _is_float(lat) and _is_float(long):
                return "POINT(" + long + " " + lat + ")"
    return np.NaN


@runtime_dependency(module="geopandas", install_from=OptionalDependency.GEO)
def _plot_gis_scatter(df: pd.DataFrame, lon: str, lat: str):
    """
    Parameters
    ----------
    df : pd.DataFrame
        dataframe contains gis data
    lon: str
        column name mapping to longitude
    lat: str
        column name mapping to latitude
    Returns
    -------
        matplotlib.axes._subplots.AxesSubplot present location of given data on world map
    """
    if len(df.index):
        if lon in df.columns and lat in df.columns:
            fig, ax = plt.subplots(facecolor=SchemeNeutral.BACKGROUND_LIGHT)
            gdf = geopandas.GeoDataFrame(
                df, geometry=geopandas.points_from_xy(df[lon], df[lat])
            )
            world = geopandas.read_file(
                geopandas.datasets.get_path("naturalearth_lowres")
            )
            world = world[world["name"] == "United States of America"]
            ax1 = world.plot(
                ax=ax, color=SchemeNeutral.AREA_LIGHT, linewidth=0.5, edgecolor="white"
            )
            gdf.plot(ax=ax1, color=SchemeTeal.LINE_LIGHT, markersize=10)
            plt.axis("off")
            return ax1


def _to_lat_long(x: pd.Series, zipcode: pd.DataFrame):
    """
    Parameters
    ----------
    x: pd.Series
        pandas series contains zip code data
    zipcode : pd.DataFrame
        dataframe contains gis data of zip code
    Returns
    -------
    pd.DataFrame
        dataframe with 2 columes including latitude and longitude data
    """
    lats = []
    longs = []
    for s in x:
        if s in zipcode.index:
            lats.append(np.float64(zipcode.loc[s]["latitude"]))
            longs.append(np.float64(zipcode.loc[s]["longitude"]))
    df = pd.DataFrame(list(zip(lats, longs)), columns=["latitude", "longitude"])
    return df


@runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
def _set_seaborn_theme():
    """
    Sets seaborn figure & axes facecolor
    """
    plt.figure(facecolor=SchemeNeutral.BACKGROUND_LIGHT)
    seaborn.set(
        rc={
            "axes.facecolor": SchemeNeutral.BACKGROUND_LIGHT,
            "figure.facecolor": SchemeNeutral.BACKGROUND_LIGHT,
        }
    )


def _format_stat(stat: pd.Series):
    """
    Formats statistics row index.
    """
    stat.rename(
        {
            "std": "standard deviation",
            "min": "sample minimum",
            "25%": "lower quartile",
            "50%": "median",
            "75%": "upper quartile",
            "max": "sample maximum",
        },
        inplace=True,
    )


@lru_cache(maxsize=1)
def _zip_code():
    """Returns dataframe contain zip code, latitude and longitude data."""
    return pd.DataFrame.from_dict(
        zip_code_dict, orient="index", columns=["longitude", "latitude"]
    )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import warnings

warnings.warn(
    (
        "The `ads.catalog.project` is deprecated in `oracle-ads 2.6.9` and will be removed in `oracle-ads 3.0`."
    ),
    DeprecationWarning,
    stacklevel=2,
)

from ads.catalog.summary import SummaryList
from ads.common import oci_client, auth
from ads.common import utils
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.config import (
    OCI_ODSC_SERVICE_ENDPOINT,
    OCI_IDENTITY_SERVICE_ENDPOINT,
    NB_SESSION_COMPARTMENT_OCID,
)
from collections.abc import Mapping
from oci.config import from_file
from oci.data_science.models import (
    Project,
    ProjectSummary,
    CreateProjectDetails,
    UpdateProjectDetails,
)
from oci.exceptions import ServiceError
from pandas import DataFrame
from types import MethodType


create_project_details_attributes = CreateProjectDetails().swagger_types.keys()
update_project_details_attributes = UpdateProjectDetails().swagger_types.keys()
project_attributes = list(Project().swagger_types.keys())
project_attributes.append("user_name")


class ProjectSummaryList(SummaryList):
    """
    A class used to represent Project Summary List.

    ...

    Attributes
    ----------
    df : data frame
        Summary information for a project.
    datetime_format: str
        Format used to describe time.
    response : oci.response.Response
        A response object with data of type list of ProjectSummaryList.
    short_id_index: (dict of str: str)
        Mapping of short id and its value.

    Methods
    -------
    sort_by(self, columns, reverse=False):
        Sort ProjectSummaryList by columns.

    filter(self, selection, instance=None):
        Filter the project list according to a lambda filter function, or list comprehension.
    """

    def __init__(self, project_list, response=None, datetime_format=utils.date_format):
        super(ProjectSummaryList, self).__init__(
            project_list, datetime_format=datetime_format
        )
        self.response = response

    def __add__(self, rhs):
        return ProjectSummaryList(
            list.__add__(self, rhs), datetime_format=self.datetime_format
        )

    def sort_by(self, columns, reverse=False):
        """Sort ProjectSummaryList by columns.

        Performs a multi-key sort on a particular set of columns and returns the sorted ProjectSummaryList
        Results are listed in a descending order by default.

        Parameters
        ----------
        columns: List of string
          A list of columns which are provided to sort on
        reverse: Boolean (defaults to false)
          If you'd like to reverse the results (for example, to get ascending instead of descending results)

        Returns
        -------
        ProjectSummaryList: A sorted ProjectSummaryList
        """
        return ProjectSummaryList(
            self._sort_by(columns, reverse=reverse),
            datetime_format=self.datetime_format,
        )

    def filter(self, selection, instance=None):
        """Filter the project list according to a lambda filter function, or list comprehension.

        Parameters
        ----------
        selection: lambda function filtering Project instances, or a list-comprehension
            function of list filtering projects
        instance: list, optional
            list to filter, optional, defaults to self

        Returns
        -------
        ProjectSummaryList: A filtered ProjectSummaryList

        Raises
        -------
        ValueError: If selection passed is not correct.
        """
        instance = instance if instance is not None else self

        if callable(selection):
            res = list(filter(selection, instance))
            # lambda filtering
            if len(res) == 0:
                print("No project found")
                return
            return ProjectSummaryList(res)
        elif isinstance(selection, list):
            # list comprehension
            if len(selection) == 0:
                print("No project found")
                return
            return ProjectSummaryList(selection, datetime_format=self.datetime_format)
        else:
            raise ValueError(
                "Filter selection must be a function or a ProjectSummaryList"
            )


class ProjectCatalog(Mapping):
    def __init__(
        self, compartment_id=None, ds_client_auth=None, identity_client_auth=None
    ):
        self.compartment_id = (
            NB_SESSION_COMPARTMENT_OCID if compartment_id is None else compartment_id
        )
        if self.compartment_id is None:
            raise ValueError("compartment_id is required")

        self.ds_client_auth = ds_client_auth or auth.default_signer(
            {"service_endpoint": OCI_ODSC_SERVICE_ENDPOINT}
            if OCI_ODSC_SERVICE_ENDPOINT
            else None
        )
        self.identity_client_auth = identity_client_auth or auth.default_signer(
            {"service_endpoint": OCI_IDENTITY_SERVICE_ENDPOINT}
            if OCI_IDENTITY_SERVICE_ENDPOINT
            else None
        )

        self.ds_client = oci_client.OCIClientFactory(**self.ds_client_auth).data_science
        self.identity_client = oci_client.OCIClientFactory(
            **self.identity_client_auth
        ).identity
        self.short_id_index = {}

    def __getitem__(self, key):
        return self.get_project(key)

    def __iter__(self):
        return self.list_projects().__iter__()

    def __len__(self):
        return len(self.list_projects())

    def _decorate_project(self, project, response=None):
        project.catalog = self
        project.response = response
        project.user_name = ""
        project.swagger_types["user_name"] = "str"
        project.swagger_types["user_email"] = "str"
        project.swagger_types["short_id"] = "str"
        project.ocid = project.id
        try:
            user = self.identity_client.get_user(project.created_by)
            project.user = user.data
            project.user_name = user.data.name
            project.user_email = user.data.email
        except:
            pass

        def commit(project_self, **kwargs):
            update_project_details = UpdateProjectDetails(
                **{
                    key: getattr(project, key)
                    for key in update_project_details_attributes
                }
            )
            return project_self.catalog.update_project(
                project_self.id, update_project_details, **kwargs
            )

        def rollback(project_self):
            """
            Get back the project to remote state

            Returns
            -------
            The project from remote state
            """
            project_self.__dict__.update(self.get_project(project_self.id).__dict__)

        def to_dataframe(project_self):
            df = DataFrame.from_dict(
                {key: getattr(project_self, key) for key in project_attributes},
                orient="index",
                columns=[""],
            )
            return df

        @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
        def show_in_notebook(project_self):
            """
            Describe the project by showing it's properties
            """
            from IPython.core.display import display

            display(project_self)

        def _repr_html_(project_self):
            return (
                project_self.to_dataframe()
                .style.set_properties(**{"margin-left": "0px"})
                .render()
            )

        project.commit = MethodType(commit, project)
        project.rollback = MethodType(rollback, project)
        project.to_dataframe = MethodType(to_dataframe, project)
        project.show_in_notebook = MethodType(show_in_notebook, project)
        project._repr_html_ = MethodType(_repr_html_, project)

        return project

    def list_projects(
        self, include_deleted=False, datetime_format=utils.date_format, **kwargs
    ):
        """
        List all projects in a given compartment, or in the current notebook session's compartment

        Parameters
        ----------
        include_deleted: bool, optional, default=False
            Whether to include deleted projects in the returned list
        datetime_format: str, optional, default: '%Y-%m-%d %H:%M:%S'
            Change format for date time fields

        Returns
        -------
        ProjectSummaryList: List of Projects.

        Raises
        -------
        KeyError: If the resource was not found or do not have authorization to access that resource.
        """
        try:
            list_projects_response = self.ds_client.list_projects(
                self.compartment_id, **kwargs
            )
            if (
                list_projects_response.data is None
                or len(list_projects_response.data) == 0
            ):
                print("No project found.")
                return
        except ServiceError as se:
            if se.status == 404:
                raise KeyError(se.message) from se
            else:
                raise

        project_list_filtered = [
            self._decorate_project(project)
            for project in list_projects_response.data
            if include_deleted
            or project.lifecycle_state != ProjectSummary.LIFECYCLE_STATE_DELETED
        ]

        # handle empty list
        if project_list_filtered is None or len(project_list_filtered) == 0:
            print("No project found.")
            return []

        psl = ProjectSummaryList(
            project_list_filtered,
            list_projects_response,
            datetime_format=datetime_format,
        )
        self.short_id_index.update(psl.short_id_index)
        return psl

    def create_project(self, create_project_details=None, **kwargs):
        """
        Create a new project with the supplied details.
        create_project_details contains parameters needed to create a new project, according to oci.data_science.models.CreateProjectDetails.

        Parameters
        ----------
        display_name: str
            The value to assign to the display_name property of this CreateProjectDetails.
        description: str
            The value to assign to the description property of this CreateProjectDetails.
        compartment_id: str
            The value to assign to the compartment_id property of this CreateProjectDetails.
        freeform_tags: dict(str, str)
            The value to assign to the freeform_tags property of this CreateProjectDetails.
        defined_tags: dict(str, dict(str, object))
            The value to assign to the defined_tags property of this CreateProjectDetails.
        kwargs:
            New project details can be supplied instead as kwargs

        Returns
        -------
        oci.data_science.models.Project: A new Project record.
        """
        if create_project_details is None:
            create_project_details = CreateProjectDetails(
                **{
                    k: v
                    for k, v in kwargs.items()
                    if k in create_project_details_attributes
                }
            )
            create_project_details.compartment_id = self.compartment_id
            # filter kwargs removing used keys
            kwargs = {
                k: v
                for k, v in kwargs.items()
                if k not in create_project_details_attributes
            }
        try:
            create_project_response = self.ds_client.create_project(
                create_project_details, **kwargs
            )
        except ServiceError as se:
            if se.status == 404:
                raise KeyError(se.message) from se
            else:
                raise
        return self._decorate_project(
            create_project_response.data, response=create_project_response
        )

    def get_project(self, project_id):
        """
        Get the Project based on project_id

        Parameters
        ----------
        project_id: str, required
            The OCID of the project to get.

        Returns
        -------
        The oci.data_science.models.Project with the matching ID.

        Raises
        -------
        KeyError: If the resource was not found or do not have authorization to access that resource.
        """
        if not project_id.startswith("ocid"):
            project_id = self.short_id_index[project_id]

        try:
            get_project_response = self.ds_client.get_project(project_id)
        except ServiceError as se:
            if se.status == 404:
                raise KeyError(se.message) from se
            else:
                raise
        return self._decorate_project(get_project_response.data, get_project_response)

    def update_project(self, project_id, update_project_details=None, **kwargs):
        """
        Updates a project with given project_id, using the provided update data
        update_project_details contains the update project details data to apply, according to oci.data_science.models.UpdateProjectDetails

        Parameters
        ----------
        project_id: str
            project_id OCID to update
        display_name: str
            The value to assign to the display_name property of this UpdateProjectDetails.
        description: str
            The value to assign to the description property of this UpdateProjectDetails.
        freeform_tags: dict(str, str)
            The value to assign to the freeform_tags property of this UpdateProjectDetails.
        defined_tags: dict(str, dict(str, object))
            The value to assign to the defined_tags property of this UpdateProjectDetails.
        kwargs: dict, optional
            Update project details can be supplied instead as kwargs

        Returns
        -------
        oci.data_science.models.Project: The updated Project record
        """
        if not project_id.startswith("ocid"):
            project_id = self.short_id_index[project_id]
        if update_project_details is None:
            update_project_details = UpdateProjectDetails(
                **{
                    k: v
                    for k, v in kwargs.items()
                    if k in update_project_details_attributes
                }
            )
            update_project_details.compartment_id = self.compartment_id
            # filter kwargs removing used keys
            kwargs = {
                k: v
                for k, v in kwargs.items()
                if k not in update_project_details_attributes
            }
        try:
            update_project_response = self.ds_client.update_project(
                project_id, update_project_details, **kwargs
            )
        except ServiceError as se:
            if se.status == 404:
                raise KeyError(se.message) from se
            else:
                raise
        return self._decorate_project(
            update_project_response.data, update_project_response
        )

    def delete_project(self, project, **kwargs):
        """
        Deletes the project based on project_id.

        Parameters
        ----------
        project: str ID or oci.data_science.models.Project,required
            The OCID of the project to delete as a string, or a Project instance

        Returns
        -------
        Bool: True if delete was succesful
        """

        try:
            project_id = (
                project.id
                if isinstance(project, Project)
                else self.short_id_index[project]
                if not project.startswith("ocid")
                else project
            )
            self.ds_client.delete_project(project_id, **kwargs)
            return True
        except:
            return False



#!/usr/bin/env python
# -*- coding: utf-8; -*--

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging

logger = logging.getLogger(__name__)
try:
    from ads.jobs.builders.runtimes.container_runtime import ContainerRuntime
    from ads.jobs.builders.runtimes.python_runtime import (
        GitPythonRuntime,
        NotebookRuntime,
        PythonRuntime,
        ScriptRuntime,
    )
    from ads.pipeline.ads_pipeline import Pipeline
    from ads.pipeline.ads_pipeline_run import (
        PIPELINE_RUN_TERMINAL_STATE,
        LogType,
        PipelineRun,
        LogType,
        ShowMode,
    )
    from ads.pipeline.ads_pipeline_step import PipelineStep
    from ads.pipeline.builders import infrastructure
    from ads.pipeline.builders.infrastructure.custom_script import CustomScriptStep
    from ads.pipeline.visualizer.base import GraphOrientation

except AttributeError as e:
    import oci.data_science

    if not hasattr(oci.data_science.models, "Pipeline"):
        logger.warning(
            "The OCI SDK you installed does not support Data Science Pipleines. ADS Pipelines API will not work."
        )
    else:
        raise e



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from cycler import cycler
import logging
import matplotlib as mpl
import numpy as np
from numpy.typing import ArrayLike
import pandas as pd
import re
from sklearn.preprocessing import LabelEncoder
import tempfile
from typing import List, Any

logging.getLogger("matplotlib").setLevel(logging.WARNING)
mpl.rcParams["image.cmap"] = "BuGn"
mpl.rcParams["axes.prop_cycle"] = cycler(
    color=["teal", "blueviolet", "forestgreen", "peru", "y", "dodgerblue", "r"]
)

from ads.common.data import ADSData
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common.decorator.deprecate import deprecated
from ads.common import logger
from ads.common.model import ADSModel
from ads.common.model_metadata import UseCaseType
from ads.dataset.dataset_with_target import ADSDatasetWithTarget
from ads.evaluations.evaluation_plot import EvaluationPlot
from ads.evaluations.statistical_metrics import (
    ModelEvaluator,
    DEFAULT_BIN_CLASS_METRICS,
    DEFAULT_MULTI_CLASS_METRICS,
    DEFAULT_REG_METRICS,
    DEFAULT_BIN_CLASS_LABELS_MAP,
    DEFAULT_MULTI_CLASS_LABELS_MAP,
    DEFAULT_REG_LABELS_MAP,
)
from ads.model.generic_model import GenericModel, VERIFY_STATUS_NAME

METRICS_TO_MINIMIZE = ["hamming_loss", "hinge_loss", "mse", "mae"]
POSITIVE_CLASS_NAMES = ["yes", "y", "t", "true", "1"]


class Evaluator(object):
    """
    BETA FEATURE
    Evaluator is the new and preferred way to evaluate a model of list of models.
    It contains a superset of the features of the soon-to-be-deprecated ADSEvaluator.

    Methods
    -------
    display()
        Shows all plots and metrics within the jupyter notebook.
    html()
        Returns the raw string of the html report
    save(filename)
        Saves the html report to the provided file location.
    add_model(model)
        Adds a model to the existsing report. See documentation for more details.
    add_metric(metric_fn)
        Adds a metric to the existsing report. See documentation for more details.
    add_plot(plotting_fn)
        Adds a plot to the existing report. See documentation for more details.

    """

    def __init__(
        self,
        models: List[GenericModel],
        X: ArrayLike,
        y: ArrayLike,
        y_preds: List[ArrayLike] = None,
        y_scores: List[ArrayLike] = None,
        X_train: ArrayLike = None,
        y_train: ArrayLike = None,
        classes: List = None,
        positive_class: str = None,
        legend_labels: dict = None,
        use_case_type: UseCaseType = None,
    ):
        """Creates an ads evaluator object.

        Parameters
        ----------
        models : ads.model.GenericModel instance
            Test data to evaluate model on.
            The object can be built using from one of the framworks supported in `ads.model.framework`
        X : DataFrame-like
            The data used to make a prediction.
            Can be set to None if `y_preds` is given. (And `y_scores` for more thorough analysis).
        y : array-like
            The true values corresponding to the input data
        y_preds : list of array-like, optional
            The predictions from each model in the same order as the models
        y_scores : list of array-like, optional
            The predict_probas from each model in the same order as the models
        X_train : DataFrame-like, optional
            The data used to train the model
        y_train : array-like, optional
            The true values corresponding to the input training data
        positive_class : str or int, optional
            The class to report metrics for binary dataset. If the target classes is True or False,
            positive_class will be set to True by default. If the dataset is multiclass or multilabel,
            this will be ignored.
        legend_labels : dict, optional
            List of legend labels. Defaults to `None`.
            If legend_labels not specified class names will be used for plots.
        classes : List or None, optional
            A List of the possible labels for y, when evaluating a classification use case
        use_case_type : str, optional
            The type of problem this model is solving. This can be set during `prepare()`.
            Examples: "binary_classification", "regression", "multinomial_classification"
            Full list of supported types can be found here: `ads.common.model_metadata.UseCaseType`

        Examples
        --------
        >>> import tempfile
        >>> from ads.evaluations.evaluator import Evaluator
        >>> from sklearn.tree import DecisionTreeClassifier
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.model_selection import train_test_split
        >>> from ads.model.framework.sklearn_model import SklearnModel
        >>> from ads.common.model_metadata import UseCaseType
        >>>
        >>> X, y = make_classification(n_samples=1000)
        >>> X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3)
        >>> est = DecisionTreeClassifier().fit(X_train, y_train)
        >>> model = SklearnModel(estimator=est, artifact_dir=tempfile.mkdtemp())
        >>> model.prepare(
                inference_conda_env="generalml_p38_cpu_v1",
                training_conda_env="generalml_p38_cpu_v1",
                X_sample=X_test,
                y_sample=y_test,
                use_case_type=UseCaseType.BINARY_CLASSIFICATION,
            )
        >>> report = Evaluator([my_model], X=X_test, y=y_test)
        >>> report.display()

        """
        self._verify_models(models)
        self.X, self.y, self.X_train, self.y_train = X, y, X_train, y_train
        self.legend_labels = legend_labels
        self.positive_class = positive_class

        self._determine_problem_type(models, use_case_type)
        self._determine_classes(classes)

        self.model_names = []
        self.evaluation = pd.DataFrame()
        self.add_models(models, y_preds=y_preds, y_scores=y_scores)

    def _verify_models(self, models):
        assert isinstance(
            models, list
        ), f"The `models` argument must be a list of models, instead got: {models}"
        for m in models:
            if not isinstance(m, GenericModel):
                raise ValueError(
                    f"Please register and prepare model {m} with ads. More information here: https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/introduction.html#register"
                )
            sum_stat = m.summary_status().reset_index()
            if (
                sum_stat.loc[sum_stat["Step"] == VERIFY_STATUS_NAME, "Status"]
                == "Not Available"
            ).any():
                raise ValueError(
                    f"Model {m} has not been prepared, and `verify` cannot be run (including the pre and post processing from the score.py). This may cause issues. Prepare the model in accordance with the documentation: https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/model_artifact.html#prepare-the-model-artifact"
                )

    def _determine_problem_type(self, models, use_case_type):
        if use_case_type is not None:
            self.problem_type = use_case_type
        problem_type = models[0].metadata_taxonomy["UseCaseType"].value
        if problem_type is not None:
            for m in models:
                assert (
                    problem_type == m.metadata_taxonomy["UseCaseType"].value
                ), f"Cannot compare models of different Use Case types. The first model is of type {problem_type}, while model: {m} is of Use Case type: {m.metadata_taxonomy['UseCaseType'].value}"
            self.problem_type = problem_type
        else:
            if not models[0].schema_output.keys:
                raise ValueError(
                    f"The Use Case Type of this model, {models[0]}, is ambigious. Please re-run Evaluator with `use_case_type` set to a valid type (full list found here: ads.common.model_metadata.UseCaseType). To avoid setting this in the future, set the `use_case_type` when preparing the model. Or update your model's use_case_type attribute here: `model.metadata_taxonomy['UseCaseType'].value` More information here: https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/model_metadata.html#taxonomy-metadata"
                )
            logger.warn(
                f"The Use Case Type of this model, {models[0]}, is ambigious. Please set the `model.metadata_taxonomy['UseCaseType'].value` attribute to one of the options in ads.common.model_metadata.UseCaseType"
            )

            output_col = models[0].schema_output.keys[0]
            if models[0].schema_output[output_col].feature_type != "Continuous":
                if len(np.unique(self.y)) == 2:
                    self.problem_type = UseCaseType.BINARY_CLASSIFICATION
                else:
                    self.problem_type = UseCaseType.MULTINOMIAL_CLASSIFICATION
            else:
                self.problem_type = UseCaseType.REGRESSION
            logger.info(f"Set Use Case Type to: {self.problem_type}")

    def _determine_classes(self, classes):
        if self.problem_type in [UseCaseType.REGRESSION]:
            self.classes = []
            self.metrics_to_show = DEFAULT_REG_METRICS
            self.is_classifier = False
        else:
            self.is_classifier = True
            self.classes = (
                classes or np.unique(self.y_train)
                if self.y_train is not None
                else np.unique(self.y)
            )
            self.num_classes = len(self.classes)
            if len(self.classes) == 2:
                self.metrics_to_show = DEFAULT_BIN_CLASS_METRICS
                if (
                    self.positive_class is None
                    or self.positive_class not in self.classes
                ):
                    self.positive_class = next(
                        (
                            x
                            for x in list(self.classes)
                            if str(x).lower() in POSITIVE_CLASS_NAMES
                        ),
                        self.classes[0],
                    )
                    logger.info(
                        f"Using {self.positive_class} as the positive class. Use `positive_class` to set this value."
                    )
            else:
                self.metrics_to_show = DEFAULT_MULTI_CLASS_METRICS

    def _get_model_name(self, model):
        name = str(model.algorithm) + "_" + str(model.framework)
        name_edit = re.sub(r" ?\([^)]+\)", "", name)
        if name_edit in self.model_names:
            name_edit += "_1"
            num_tries = 1
            while name_edit in self.model_names:
                num_tries += 1
                name_edit = name_edit[:-1] + str(num_tries)
        self.model_names.append(name_edit)
        return name_edit

    def _score_data(self, model, X):
        y_pred = model.verify(X)["prediction"]

        y_score = None
        # we will compute y_score only for binary classification cases because only for binary classification can
        # we use it for ROC Curves and AUC
        if self.is_classifier and hasattr(model.estimator, "predict_proba"):
            if len(self.classes) == 2:
                #  positive label index is assumed to be 0 if the ADSModel does not have a positive class defined
                positive_class_index = 0
                # For prediction probability, we only consider the positive class.
                if self.positive_class is not None:
                    if self.positive_class not in list(self.classes):
                        raise ValueError(
                            "Invalid positive class '%s' for model %s. Positive class should be one of %s."
                            % (
                                self.positive_class,
                                model.estimator.__class__.__name__,
                                list(self.classes),
                            )
                        )
                    positive_class_index = list(self.classes).index(self.positive_class)
                y_score = model.estimator.predict_proba(X)[:, positive_class_index]
            else:
                y_score = model.estimator.predict_proba(X)
        return y_pred, y_score

    def add_models(
        self,
        models: List[GenericModel],
        y_preds: List[Any] = None,
        y_scores: List[Any] = None,
    ):
        """Add a model to an existing Evaluator to avoid re-calculating the values.

        Parameters
        ----------
        models : List[ads.model.GenericModel]
            Test data to evaluate model on.
            The object can be built using from one of the framworks supported in `ads.model.framework`
        y_preds : list of array-like, optional
            The predictions from each model in the same order as the models
        y_scores : list of array-like, optional
            The predict_probas from each model in the same order as the models

        Returns
        -------
        self

        Examples
        --------
        >>> evaluator = Evaluator(models = [model1, model2], X=X, y=y)
        >>> evaluator.add_models(models = [model3])
        """

        assert isinstance(models, List), "The `models` parameter must be of type list."
        if self.is_classifier:
            self._le = LabelEncoder().fit(self.y)
        for i, m in enumerate(models):
            m_name = self._get_model_name(m)

            if y_preds is None:
                y_pred, y_score = self._score_data(m, self.X)
            else:
                y_pred = y_preds[i]
                y_score = y_scores[i] if isinstance(y_scores, list) else None
            if self.is_classifier:
                y_true, y_pred = self._le.transform(self.y), self._le.transform(y_pred)
                classes = self._le.transform(self.classes)
                pos_class = None
                if len(self.classes) == 2:
                    pos_class = self._le.transform([self.positive_class])[0]
            else:
                y_true, y_pred, classes, pos_class = self.y, y_pred, None, None
            new_model_metrics = ModelEvaluator(
                y_true=y_true,
                y_pred=y_pred,
                model_name=m_name,
                classes=classes,
                y_score=y_score,
                positive_class=pos_class,
            ).get_metrics()
            self.evaluation = pd.concat(
                [self.evaluation, new_model_metrics], axis=1, sort=False
            )
        return self

    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    @runtime_dependency(
        module="ipywidgets",
        object="HTML",
        install_from=OptionalDependency.NOTEBOOK,
    )
    def display(
        self,
        plots=None,
        perfect=False,
        baseline=True,
        legend_labels=None,
        precision=4,
        metrics_labels=None,
    ):
        """Visualize evaluation report.

        Parameters
        ----------
        plots : list, optional
            Filter the plots that are displayed. Defaults to None. The name of the plots are as below:

                - regression - residuals_qq, residuals_vs_fitted
                - binary classification - normalized_confusion_matrix, roc_curve, pr_curve
                - multi class classification - normalized_confusion_matrix, precision_by_label, recall_by_label, f1_by_label
        perfect: bool, optional (default False)
            If True, will show how a perfect classifier would perform.
        baseline: bool, optional (default True)
            If True, will show how a random classifier would perform.
        legend_labels : dict, optional
            Rename legend labels, that used for multi class classification plots. Defaults to None.
            legend_labels dict keys are the same as class names. legend_labels dict values are strings.
            If legend_labels not specified class names will be used for plots.
        precision: int, optional (default 4)
            The number of decimal points to show for each score/loss value
        metrics_labels: List, optional
            The metrics that should be included in the html table.

        Returns
        -------
        None
            Nothing. Outputs several evaluation plots as specified by `plots`.

        Examples
        --------

        >>> evaluator = Evaluator(models=[model1, model2], X=X, y=y)
        >>> evaluator.display()

        >>> legend_labels={'class_0': 'green', 'class_1': 'yellow', 'class_2': 'red'}
        >>> multi_evaluator = Evaluator(models=[model1, model2], X=X, y=y, legend_labels=legend_labels)
        >>> multi_evaluator.display(plots=["normalized_confusion_matrix",
        ...             "precision_by_label", "recall_by_label", "f1_by_label"])
        """
        from IPython.core.display import display, HTML

        legend_labels = (
            legend_labels if legend_labels is not None else self.legend_labels
        )
        if legend_labels is None and self.is_classifier:
            legend_labels = dict(
                zip([str(x) for x in self._le.transform(self.classes)], self.classes)
            )
        # pass to plotting class
        self._get_plots_html(
            plots=plots, perfect=perfect, baseline=baseline, legend_labels=legend_labels
        )
        display(
            HTML(self._get_metrics_html(precision=precision, labels=metrics_labels))
        )

    def html(
        self,
        plots=None,
        perfect=False,
        baseline=True,
        legend_labels=None,
        precision=4,
        metrics_labels=None,
    ):
        """Get raw HTML report.

        Parameters
        ----------
        plots : list, optional
            Filter the plots that are displayed. Defaults to None. The name of the plots are as below:

                - regression - residuals_qq, residuals_vs_fitted
                - binary classification - normalized_confusion_matrix, roc_curve, pr_curve
                - multi class classification - normalized_confusion_matrix, precision_by_label, recall_by_label, f1_by_label
        perfect: bool, optional (default False)
            If True, will show how a perfect classifier would perform.
        baseline: bool, optional (default True)
            If True, will show how a random classifier would perform.
        legend_labels : dict, optional
            Rename legend labels, that used for multi class classification plots. Defaults to None.
            legend_labels dict keys are the same as class names. legend_labels dict values are strings.
            If legend_labels not specified class names will be used for plots.
        precision: int, optional (default 4)
            The number of decimal points to show for each score/loss value
        metrics_labels: List, optional
            The metrics that should be included in the html table.

        Returns
        -------
        None
            Nothing. Outputs several evaluation plots as specified by `plots`.

        Examples
        --------

        >>> evaluator = Evaluator(models=[model1, model2], X=X, y=y)
        >>> raw_html = evaluator.html()
        """
        html_plots = self._get_plots_html(
            plots=plots, perfect=perfect, baseline=baseline, legend_labels=legend_labels
        )
        html_metrics = self._get_metrics_html(
            precision=precision, labels=metrics_labels
        )
        html_raw = (
            "<h1>Evaluation Report</h1> \
                    <h2>Evaluation Plots</h2> "
            + " \
                    ".join(
                html_plots
            )
            + f"<h2>Evaluation Metrics</h2>  \
                    <p> {html_metrics} </p>"
        )
        return html_raw

    def save(self, filename: str, **kwargs):
        """Save HTML report.

        Parameters
        ----------
        filename: str
            The name and path of where to save the html report.
        plots : list, optional
            Filter the plots that are displayed. Defaults to None. The name of the plots are as below:

                - regression - residuals_qq, residuals_vs_fitted
                - binary classification - normalized_confusion_matrix, roc_curve, pr_curve
                - multi class classification - normalized_confusion_matrix, precision_by_label, recall_by_label, f1_by_label
        perfect: bool, optional (default False)
            If True, will show how a perfect classifier would perform.
        baseline: bool, optional (default True)
            If True, will show how a random classifier would perform.
        legend_labels : dict, optional
            Rename legend labels, that used for multi class classification plots. Defaults to None.
            legend_labels dict keys are the same as class names. legend_labels dict values are strings.
            If legend_labels not specified class names will be used for plots.
        precision: int, optional (default 4)
            The number of decimal points to show for each score/loss value
        metrics_labels: List, optional
            The metrics that should be included in the html table.

        Returns
        -------
        None
            Nothing. Outputs several evaluation plots as specified by `plots`.

        Examples
        --------

        >>> evaluator = Evaluator(models=[model1, model2], X=X, y=y)
        >>> evaluator.save("report.html")
        """
        raw_html = self.html(**kwargs)
        with open(filename, "w") as f:
            f.write(raw_html)

    def _get_plots_html(
        self,
        plots=None,
        perfect=False,
        baseline=True,
        legend_labels=None,
    ):
        return EvaluationPlot.plot(
            self.evaluation, plots, len(self.classes), perfect, baseline, legend_labels
        )

    def _get_metrics_html(self, precision=4, labels=None):
        def highlight_max(s):
            """Highlight the maximum in a Series yellow.

            Parameters
            ----------
            s : series object
                the series being evaluated

            Returns
            -------
            list
                containing background color data or empty if not max
            """
            if s.name not in METRICS_TO_MINIMIZE:
                is_max = s == s.max()
            else:
                is_max = s == s.min()
            return ["background-color: lightgreen" if v else "" for v in is_max]

        def _pretty_label(df, labels, copy=True):
            """
            Output specified labels in proper format.
            If the labels are provided in then used them. Otherwise, use default.

            Parameters
            ----------
            labels : dictionary
                map printing specific labels for metrics display

            Returns
            -------
            dataframe
                dataframe with index names modified according to input labels
            """
            df_display = df.loc[list(labels.keys())]

            if copy:
                df_display = df_display.copy()
            for k, v in labels.items():
                df_display.rename(index={k: v}, inplace=True)
            return df_display

        if labels is None:
            if self.is_classifier:
                if len(self.classes) == 2:
                    labels = DEFAULT_BIN_CLASS_LABELS_MAP
                else:
                    labels = DEFAULT_MULTI_CLASS_LABELS_MAP
            else:
                labels = DEFAULT_REG_LABELS_MAP
        html_raw = (
            _pretty_label(self.evaluation, labels)
            .style.apply(highlight_max, axis=1)
            .format(precision=precision)
            .set_properties(**{"text-align": "center"})
            .set_table_attributes("class=table")
            .set_caption(
                '<div align="left"><b style="font-size:20px;">'
                + "Evaluation Metrics:</b></div>"
            )
            .to_html()
        )
        return html_raw


class ADSEvaluator(object):
    """ADS Evaluator class. This class holds field and methods for creating and using
    ADS evaluator objects.

    Attributes
    ----------
    evaluations : list[DataFrame]
        list of evaluations.
    is_classifier : bool
        Whether the dataset looks like a classification problem (versus regression).
    legend_labels : dict
        List of legend labels. Defaults to `None`.
    metrics_to_show : list[str]
        Names of metrics to show.
    models : list[ads.common.model.ADSModel]
        The object built using `ADSModel.from_estimator()`.
    positive_class : str or int
        The class to report metrics for binary dataset, assumed to be true.
    show_full_name :bool
        Whether to show the name of the evaluator in relevant contexts.
    test_data : ads.common.data.ADSData
        Test data to evaluate model on.
    training_data : ads.common.data.ADSData
        Training data to evaluate model.

    Positive_Class_names : list
        Class attribute listing the ways to represent positive classes

    Methods
    -------
    add_metrics(func, names)
        Adds the listed metics to the evaluator it is called on
    del_metrics(names)
        Removes listed metrics from the evaluator object it is called on
    add_models(models, show_full_name)
        Adds the listed models to the evaluator object
    del_models(names)
        Removes the listed models from the evaluator object
    show_in_notebook(plots, use_training_data, perfect, baseline, legend_labels)
        Visualize evalutation plots in the notebook
    calculate_cost(tn_weight, fp_weight, fn_weight, tp_weight, use_training_data)
        Returns a cost associated with the input weights
    """

    Positive_Class_Names = ["yes", "y", "t", "true", "1"]

    def __init__(
        self,
        test_data,
        models,
        training_data=None,
        positive_class=None,
        legend_labels=None,
        show_full_name=False,
        classes=None,
        classification_threshold=50,
    ):
        """Creates an ads evaluator object.

        Parameters
        ----------
        test_data : ads.common.data.ADSData instance
            Test data to evaluate model on.
            The object can be built using `ADSData.build()`.
        models : list[ads.common.model.ADSModel]
            The object can be built using `ADSModel.from_estimator()`.
            Maximum length of the list is 3
        training_data : ads.common.data.ADSData instance, optional
            Training data to evaluate model on and compare metrics against test data.
            The object can be built using `ADSData.build()`
        positive_class : str or int, optional
            The class to report metrics for binary dataset. If the target classes is True or False,
            positive_class will be set to True by default. If the dataset is multiclass or multilabel,
            this will be ignored.
        legend_labels : dict, optional
            List of legend labels. Defaults to `None`.
            If legend_labels not specified class names will be used for plots.
        show_full_name : bool, optional
            Show the name of the evaluator object. Defaults to `False`.
        classes : List or None, optional
            A List of the possible labels for y, when evaluating a classification use case
        classification_threshold : int, defaults to 50
            The maximum number of unique values that y must have to qualify as classification.
            If this threshold is exceeded, Evaluator assumes the model is regression.

        Examples
        --------

        >>> train, test = ds.train_test_split()
        >>> model1 = MyModelClass1.train(train)
        >>> model2 = MyModelClass2.train(train)
        >>> evaluator = ADSEvaluator(test, [model1, model2])

        >>> legend_labels={'class_0': 'one', 'class_1': 'two', 'class_2': 'three'}
        >>> multi_evaluator = ADSEvaluator(test, models=[model1, model2],
        ...             legend_labels=legend_labels)

        """
        if any(isinstance(m, ADSModel) for m in models):
            logger.warn(
                f"ADSModel is being deprecated. Users should instead use GenericModel or one of its subclasses. More information here: https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/introduction.html#register"
            )
        self.evaluations = []
        if isinstance(training_data, ADSDatasetWithTarget):
            training_data, _ = training_data.train_test_split(test_size=0.0)
        if isinstance(test_data, ADSDatasetWithTarget):
            test_data, _ = test_data.train_test_split(test_size=0.0)

        if not isinstance(test_data, ADSData):
            raise ValueError(
                "Expected test_data to be of type ADSData. More information here: https://accelerated-data-science.readthedocs.io/en/latest/ads.common.html#ads.common.data.ADSData"
            )
        if training_data and not isinstance(training_data, ADSData):
            raise ValueError(
                "Expected training_data to be of type ADSData. More information here: https://accelerated-data-science.readthedocs.io/en/latest/ads.common.html#ads.common.data.ADSData"
            )
        assert isinstance(
            models, list
        ), "The `models` argument should be a list of GenericModels. More information here: https://accelerated-data-science.readthedocs.io/en/latest/ads.common.html#ads.common.data.ADSData"

        self.test_data = test_data
        self.training_data = training_data
        self.classes = []
        self.is_classifier = (
            hasattr(models[0], "classes_") and models[0].classes_ is not None
        )
        pclass = positive_class
        if self.is_classifier:
            self.classes = list(models[0].classes_)
            if len(self.classes) == 2:
                self.metrics_to_show = [
                    "accuracy",
                    "hamming_loss",
                    "precision",
                    "recall",
                    "f1",
                    "auc",
                ]
                if positive_class is None or positive_class not in self.classes:
                    pclass = next(
                        (
                            x
                            for x in list(self.classes)
                            if str(x).lower() in ADSEvaluator.Positive_Class_Names
                        ),
                        self.classes[0],
                    )
                    logger.info(
                        f"Using {pclass} as the positive class. Use `positive_class` to set this value."
                    )
            else:
                # Multi-class
                self.metrics_to_show = [
                    "accuracy",
                    "hamming_loss",
                    "precision_weighted",
                    "precision_micro",
                    "recall_weighted",
                    "recall_micro",
                    "f1_weighted",
                    "f1_micro",
                ]
        else:
            # Regression
            self.metrics_to_show = ["r2_score", "mse", "mae"]
        self.positive_class = pclass
        self.legend_labels = legend_labels

        for m in models:
            if not (isinstance(m, ADSModel)):
                try:
                    m = ADSModel.from_estimator(m.est)
                except:
                    logger.info("This model cannot be converted to an ADS Model.")
        self.evaluations = [pd.DataFrame(), pd.DataFrame()]
        self.model_names = []
        self.add_models(models, show_full_name=show_full_name)

    def add_metrics(self, funcs, names):
        """Adds the listed metrics to the evaluator object it is called on.

        Parameters
        ----------
        funcs : list
            The list of metrics to be added. This function will be provided `y_true`
            and `y_pred`, the true and predicted values for each model.
        names : list[str])
            The list of metric names corresponding to the functions.

        Returns
        -------
        Nothing

        Examples
        --------
        >>> def f1(y_true, y_pred):
        ...    return np.max(y_true - y_pred)
        >>> evaluator = ADSEvaluator(test, [model1, model2])
        >>> evaluator.add_metrics([f1], ['Max Residual'])
        >>> evaluator.metrics
        Output table will include the desired metric
        """

        if len(funcs) != len(names):
            raise ValueError("Could not find 1 unique name for each function")
        for name, f in zip(names, funcs):
            f_res = []
            for m in self.evaluations[1].columns:
                res = f(
                    self.evaluations[1][m]["y_true"], self.evaluations[1][m]["y_pred"]
                )
                f_res.append(res)
            pd_res = pd.DataFrame(
                [f_res], columns=self.evaluations[1].columns, index=[name]
            )
            self.evaluations[1] = pd.concat([self.evaluations[1], pd_res])
            if self.evaluations[0].shape != (0, 0):
                f_res = []
                for m in self.evaluations[0].columns:
                    res = f(
                        self.evaluations[0][m]["y_true"],
                        self.evaluations[0][m]["y_pred"],
                    )
                    f_res.append(res)
                pd_res = pd.DataFrame(
                    [f_res], columns=self.evaluations[0].columns, index=[name]
                )
                self.evaluations[0] = pd.concat([self.evaluations[0], pd_res])
            if name not in self.metrics_to_show:
                self.metrics_to_show.append(name)
        setattr(self, "train_evaluations", self.evaluations[0])
        setattr(self, "test_evaluations", self.evaluations[1])

    def del_metrics(self, names):
        """Removes the listed metrics from the evaluator object it is called on.

        Parameters
        ----------
        names : list[str]
            The list of names of metrics to be deleted. Names can be found by calling
            `evaluator.test_evaluations.index`.

        Returns
        -------
        None
            `None`

        Examples
        --------
        >>> evaluator = ADSEvaluator(test, [model1, model2])
        >>> evaluator.del_metrics(['mse])
        >>> evaluator.metrics
        Output table will exclude the desired metric
        """
        self.evaluations[1].drop(index=names, inplace=True)
        if self.evaluations[0].shape != (0, 0):
            self.evaluations[0].drop(index=names, inplace=True)
        self.metrics_to_show = [met for met in self.metrics_to_show if met not in names]

    def add_models(self, models, show_full_name=False):
        """Adds the listed models to the evaluator object it is called on.

        Parameters
        ----------
        models : list[ADSModel]
            The list of models to be added
        show_full_name : bool, optional
            Whether to show the full model name. Defaults to False.
            ** NOT USED **

        Returns
        -------
        Nothing

        Examples
        --------
        >>> evaluator = ADSEvaluator(test, [model1, model2])
        >>> evaluator.add_models("model3])
        """

        if type(models) is list:
            total_train_metrics = self.evaluations[0]
            total_test_metrics = self.evaluations[1]
            for i, m in enumerate(models):
                # if hasattr(m, 'classes_') != self.is_classifier:
                #     raise ValueError("All models should belong to same problem type.")
                # calculate evaluations on testing and training data (if X_train is not None)
                m_name = self._get_model_name(m.name)

                if self.training_data is not None:
                    y_pred, y_score = self._score_data(m, self.training_data.X)
                    train_metrics = ModelEvaluator(
                        y_true=self.training_data.y,
                        y_pred=y_pred,
                        model_name=m_name,
                        classes=m.classes_ if self.is_classifier else None,
                        y_score=y_score,
                        positive_class=self.positive_class,
                    ).get_metrics()
                    total_train_metrics = pd.concat(
                        [total_train_metrics, train_metrics], axis=1
                    )

                y_pred, y_score = self._score_data(m, self.test_data.X)
                test_metrics = ModelEvaluator(
                    y_true=self.test_data.y,
                    y_pred=y_pred,
                    model_name=m_name,
                    classes=m.classes_ if self.is_classifier else None,
                    y_score=y_score,
                    positive_class=self.positive_class,
                ).get_metrics()
                total_test_metrics = pd.concat(
                    [total_test_metrics, test_metrics], axis=1, sort=False
                )

            self.evaluations = [total_train_metrics, total_test_metrics]
            setattr(self, "train_evaluations", self.evaluations[0])
            setattr(self, "test_evaluations", self.evaluations[1])

    def del_models(self, names):
        """Removes the listed models from the evaluator object it is called on.

        Parameters
        ----------
        names : list[str]
            the list of models to be delete. Names are the model names by default, and
            assigned internally when conflicts exist. Actual names can be found using
            `evaluator.test_evaluations.columns`

        Returns
        -------
        Nothing

        Examples
        --------
        >>> model3.rename("model3")
        >>> evaluator = ADSEvaluator(test, [model1, model2, model3])
        >>> evaluator.del_models([model3])
        """

        if type(names) is list:
            self.model_names = [n for n in self.model_names if n not in names]
            self.evaluations[1].drop(columns=names, inplace=True)
            if self.evaluations[0].shape != (0, 0):
                self.evaluations[0].drop(columns=names, inplace=True)

    def show_in_notebook(
        self,
        plots=None,
        use_training_data=False,
        perfect=False,
        baseline=True,
        legend_labels=None,
    ):
        """Visualize evaluation plots.

        Parameters
        ----------
        plots : list, optional
            Filter the plots that are displayed. Defaults to None. The name of the plots are as below:

                - regression - residuals_qq, residuals_vs_fitted
                - binary classification - normalized_confusion_matrix, roc_curve, pr_curve
                - multi class classification - normalized_confusion_matrix, precision_by_label, recall_by_label, f1_by_label

        use_training_data : bool, optional
            Use training data to generate plots. Defaults to `False`.
            By default, this method uses test data to generate plots
        legend_labels : dict, optional
            Rename legend labels, that used for multi class classification plots. Defaults to None.
            legend_labels dict keys are the same as class names. legend_labels dict values are strings.
            If legend_labels not specified class names will be used for plots.

        Returns
        -------
        None
            Nothing. Outputs several evaluation plots as specified by `plots`.

        Examples
        --------

        >>> evaluator = ADSEvaluator(test, [model1, model2])
        >>> evaluator.show_in_notebook()

        >>> legend_labels={'class_0': 'green', 'class_1': 'yellow', 'class_2': 'red'}
        >>> multi_evaluator = ADSEvaluator(test, [model1, model2],
        ...             legend_labels=legend_labels)
        >>> multi_evaluator.show_in_notebook(plots=["normalized_confusion_matrix",
        ...             "precision_by_label", "recall_by_label", "f1_by_label"])
        """

        # get evaluations
        if use_training_data:
            if self.training_data is None:
                raise ValueError(
                    "Training data is not provided. Re-build ADSData with training and test data"
                )
            model_evaluation = self.evaluations[0]
        else:
            model_evaluation = self.evaluations[1]
        legend_labels = (
            legend_labels if legend_labels is not None else self.legend_labels
        )
        # pass to plotting class
        EvaluationPlot.plot(
            model_evaluation, plots, len(self.classes), perfect, baseline, legend_labels
        )

    def calculate_cost(
        self, tn_weight, fp_weight, fn_weight, tp_weight, use_training_data=False
    ):
        """Returns a cost associated with the input weights.

        Parameters
        ----------
        tn_weight : int, float
            The weight to assign true negatives in calculating the cost
        fp_weight : int, float
            The weight to assign false positives in calculating the cost
        fn_weight : int, float
            The weight to assign false negatives in calculating the cost
        tp_weight : int, float
            The weight to assign true positives in calculating the cost
        use_training_data : bool, optional
            Use training data to pull the metrics. Defaults to False

        Returns
        -------
        :class:`pandas.DataFrame`
            DataFrame with the cost calculated for each model

        Examples
        --------
        >>> evaluator = ADSEvaluator(test, [model1, model2])
        >>> costs_table = evaluator.calculate_cost(0, 10, 1000, 0)
        """

        if len(self.classes) != 2:
            raise ValueError(
                "The calculate_cost api is not supported for non-binary classification datasets."
            )
        cost_per_model = []
        if use_training_data:
            if self.training_data is None:
                raise ValueError(
                    "Training data is not provided. Re-build ADSData with training and test data."
                )
            ev = self.evaluations[0]
        else:
            ev = self.evaluations[1]
        list_of_model = ev.columns
        for m in list_of_model:
            tn, fp, fn, tp = ev[m]["raw_confusion_matrix"].ravel()
            cost_per_model.append(
                tn * tn_weight + fp * fp_weight + fn * fn_weight + tp * tp_weight
            )
        cost_df = pd.DataFrame({"model": list_of_model, "cost": cost_per_model})
        return cost_df

    class EvaluationMetrics(object):
        """Class holding evaluation metrics.

        Attributes
        ----------
        ev_test : list
            evaluation test metrics
        ev_train : list
            evaluation training metrics
        use_training : bool
            use training data
        less_is_more : list
            metrics list

        Methods
        -------
        show_in_notebook()
            Shows visualization metrics as a color coded table

        """

        DEFAULT_LABELS_MAP = {
            "accuracy": "Accuracy",
            "hamming_loss": "Hamming distance",
            "kappa_score_": "Cohen's kappa coefficient",
            "precision": "Precision",
            "recall": "Recall",
            "f1": "F1",
            "auc": "ROC AUC",
        }

        def __init__(
            self, ev_test, ev_train, use_training=False, less_is_more=None, precision=4
        ):
            self.ev_test = ev_test
            self.ev_train = ev_train
            self.use_training = use_training
            self.precision = precision
            if isinstance(less_is_more, list):
                self.less_is_more = [
                    "hamming_loss",
                    "hinge_loss",
                    "mse",
                    "mae",
                ] + less_is_more
            else:
                self.less_is_more = ["hamming_loss", "hinge_loss", "mse", "mae"]

        def __repr__(self):
            self.show_in_notebook()
            return ""

        @property
        def precision(self):
            return self._precision

        @precision.setter
        def precision(self, value):
            """
            Set precision to @property of the class.
            """
            if not isinstance(value, int):
                if not (isinstance(value, float) and value.is_integer()):
                    raise TypeError("'value' must be integer")
                value = int(value)
            if value < 0:
                raise ValueError("'value' must be non-negative")
            self._precision = value

        def show_in_notebook(self, labels=DEFAULT_LABELS_MAP):
            """
            Visualizes evaluation metrics as a color coded table.

            Parameters
            ----------
            labels : dictionary
                map printing specific labels for metrics display

            Returns
            -------
            Nothing
            """

            def highlight_max(s):
                """Highlight the maximum in a Series yellow.

                Parameters
                ----------
                s : series object
                    the series being evaluated

                Returns
                -------
                list
                    containing background color data or empty if not max
                """
                if s.name not in self.less_is_more:
                    is_max = s == s.max()
                else:
                    is_max = s == s.min()
                return ["background-color: lightgreen" if v else "" for v in is_max]

            table_styles = [
                dict(props=[("text-align", "right")]),
                dict(selector="caption", props=[("caption-side", "top")]),
            ]

            def _pretty_label(df, labels, copy=False):
                """
                Output specified labels in proper format.
                If the labels are provided in then used them. Otherwise, use default.

                Parameters
                ----------
                labels : dictionary
                    map printing specific labels for metrics display

                Returns
                -------
                dataframe
                    dataframe with index names modified according to input labels
                """
                if copy:
                    df = df.copy()
                for k, v in labels.items():
                    df.rename(index={k: v}, inplace=True)
                return df

            @runtime_dependency(
                module="IPython", install_from=OptionalDependency.NOTEBOOK
            )
            @runtime_dependency(
                module="ipywidgets",
                object="HTML",
                install_from=OptionalDependency.NOTEBOOK,
            )
            def _display_metrics(df, data_name, labels, precision):
                """
                display metrics on web page

                Parameters
                ----------
                df : dataframe
                    metrics in dataframe format
                data_name : string
                    name of data given metrics df describe
                labels : dictionary
                    map printing specific labels for metrics display
                precision : int
                    precision for metrics display

                Returns
                -------
                Nothing
                """
                from IPython.core.display import display, HTML

                display(
                    HTML(
                        _pretty_label(df, labels)
                        .style.apply(highlight_max, axis=1)
                        .format(precision=precision)
                        .set_properties(**{"text-align": "center"})
                        .set_table_attributes("class=table")
                        .set_caption(
                            '<div align="left"><b style="font-size:20px;">'
                            + "Evaluation Metrics ("
                            + data_name
                            + "):</b></div>"
                        )
                        .to_html()
                    )
                )

            _display_metrics(self.ev_test, "testing data", labels, self.precision)
            if self.use_training:
                _display_metrics(self.ev_train, "training data", labels, self.precision)

    @property
    def raw_metrics(self, metrics=None, use_training_data=False):
        """Returns the raw metric numbers

        Parameters
        ----------
        metrics : list, optional
            Request metrics to pull. Defaults to all.
        use_training_data : bool, optional
            Use training data to pull metrics. Defaults to False

        Returns
        -------
        dict
            The requested raw metrics for each model. If `metrics` is `None` return all.

        Examples
        --------
        >>> evaluator = ADSEvaluator(test, [model1, model2])
        >>> raw_metrics_dictionary = evaluator.raw_metrics()
        """

        [train_met, test_met] = self.evaluations
        test_d = test_met.to_dict()
        if use_training_data and train_met is not None:
            train_d = train_met.add_suffix("_train").to_dict()
            test_d.update(train_d)
        for m, data in test_d.items():
            ret = dict()
            for k, v in data.items():
                if isinstance(v, np.ndarray):
                    ret[k] = v.tolist()
                else:
                    ret[k] = v
            test_d[m] = ret
        return test_d

    @property
    def metrics(self):
        """Returns evaluation metrics

        Returns
        -------
        metrics
            HTML representation of a table comparing relevant metrics.

        Examples
        --------
        >>> evaluator = ADSEvaluator(test, [model1, model2])
        >>> evaluator.metrics
        Outputs table displaying metrics.
        """

        ev_test = self.evaluations[1].loc[self.metrics_to_show]
        use_training = self.evaluations[0].shape != (0, 0)
        ev_train = (
            self.evaluations[0].loc[self.metrics_to_show] if use_training else None
        )
        return ADSEvaluator.EvaluationMetrics(ev_test, ev_train, use_training)

    """
    Internal methods
    """

    def _get_model_name(self, name, show_full_name=False):
        name_edit = re.sub(r" ?\([^)]+\)", "", name)
        ## if name only has '(' without ')', the code above wouldnt remove the argument followed by '('.
        if "(" in name_edit and not show_full_name:
            name_edit = name.split("(")[0]
            logger.info("Use `show_full_name=True` to show the full model name.")
        if name_edit in self.model_names:
            name_edit += "_1"
            num_tries = 1
            while name_edit in self.model_names:
                num_tries += 1
                name_edit = name_edit[:-1] + str(num_tries)
            if num_tries == 1:
                logger.info(
                    f"The name '{name_edit[:-2]}' is used by multiple models. "
                    f"Use the `rename()` method to change the name."
                )
        self.model_names.append(name_edit)
        return name_edit

    def _score_data(self, est, X):
        y_pred = est.predict(X)
        y_score = None

        # we will compute y_score only for binary classification cases because only for binary classification can
        # we use it for ROC Curves and AUC etc
        if self.is_classifier and hasattr(est.est, "predict_proba"):
            if len(est.classes_) == 2:
                #  positive label index is assumed to be 0 if the ADSModel does not have a positive class defined
                positive_class_index = 0
                # For prediction probability, we only consider the positive class.
                if self.positive_class is not None:
                    if self.positive_class not in list(est.classes_):
                        raise ValueError(
                            "Invalid positive class '%s' for model %s. Positive class should be one of %s."
                            % (
                                self.positive_class,
                                est.est.__class__.__name__,
                                list(est.classes_),
                            )
                        )
                    positive_class_index = list(est.classes_).index(self.positive_class)
                y_score = est.predict_proba(X)[:, positive_class_index]
            else:
                y_score = est.predict_proba(X)
        return y_pred, y_score



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import subprocess
from io import StringIO
import tempfile
import os
import datetime
import shutil
import sys
import glob
import stat
import conda_pack

import yaml


def main(pack_folder_path):
    slug = os.path.basename(pack_folder_path)
    manifest_path = glob.glob(os.path.join(pack_folder_path, "*_manifest.yaml"))[0]
    with open(manifest_path) as f:
        env = yaml.safe_load(f.read())

    with tempfile.TemporaryDirectory() as td:
        process = subprocess.Popen(
            ["conda", "env", "export", "--prefix", pack_folder_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = process.communicate()
        if stderr:
            print(stderr)
            raise Exception(
                f"Error export environment information from {pack_folder_path}"
            )
        try:
            new_env_info = yaml.safe_load(StringIO(stdout.decode("utf-8")))
        except Exception as e:
            print(f"Error reading dependency list from {stdout}")
            raise e

        manifest = env["manifest"]
        manifest["type"] = "published"
        new_env_info["manifest"] = manifest
        with open(manifest_path, "w") as f:
            yaml.safe_dump(new_env_info, f)
        pack_file = os.path.join(td, f"{slug}.tar.gz")
        conda_pack.pack(
            prefix=pack_folder_path,
            compress_level=7,
            output=pack_file,
            n_threads=-1,
            ignore_missing_files=True,
        )
        if not os.path.exists(pack_file):
            raise RuntimeError(
                "Error creating the pack file using `conda_pack.pack()`."
            )
        shutil.copy(pack_file, pack_folder_path)
        file_path = os.path.join(pack_folder_path, os.path.basename(pack_file))
        print(
            f"changing permission for {file_path}",
            flush=True,
        )
        os.chmod(file_path, 0o755)


if __name__ == "__main__":
    main(sys.argv[1])



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from typing import Any, Dict, Generator, List, Tuple, Union

from ads.config import JOB_RUN_COMPARTMENT_OCID, NB_SESSION_COMPARTMENT_OCID
from ads.data_labeling.interface.loader import Loader
from ads.data_labeling.interface.parser import Parser
from ads.data_labeling.interface.reader import Reader
from ads.data_labeling.loader.file_loader import FileLoaderFactory
from ads.data_labeling.parser.dls_record_parser import DLSRecordParserFactory
from ads.data_labeling.parser.export_record_parser import RecordParserFactory
from ads.data_labeling.reader.dls_record_reader import DLSRecordReader
from ads.data_labeling.reader.export_record_reader import ExportRecordReader


class RecordReader:
    """Record Reader Class consists of parser, reader and loader. Reader reads the
    the content from the record file. Parser parses the label for each record. And
    Loader loads the content of the file path in that record.

    Examples
    --------
    >>> import os
    >>> import oci
    >>> from ads.data_labeling import RecordReader
    >>> from ads.common import auth as authutil
    >>> file_path = "/path/to/your_record.jsonl"
    >>> dataset_type = "IMAGE"
    >>> annotation_type = "BOUNDING_BOX"
    >>> record_reader = RecordReader.from_export_file(file_path, dataset_type, annotation_type, "image_file_path", authutil.api_keys())
    >>> next(record_reader.read())
    """

    def __init__(
        self,
        reader: Reader,
        parser: Parser,
        loader: Loader = None,
        include_unlabeled: bool = False,
        encoding: str = "utf-8",
        materialize: bool = False,
    ) -> "RecordReader":
        """Initiates a RecordReader instance.

        Parameters
        ----------
        reader: Reader
            Reader instance to read content from the record file.
        parser: Parser
            Parser instance to parse the labels from record file.
        loader: Loader. Defaults to None.
            Loader instance to load the content from the file path in the record.
        materialize: bool, optional. Defaults to False.
            Whether to materialize the content using loader.
        include_unlabeled: (bool, optional). Default to False.
            Whether to load the unlabeled records or not.
        encoding: str, optional
            Encoding for text files. Used only to extract the content of the text dataset contents.

        Raises
        ------
        ValueError
            If the record reader and record parser must be specified.
            If the loader is not specified when materialize if True.
        """
        if not reader:
            raise ValueError("The record reader must be specified.")
        if not parser:
            raise ValueError("The record parser must be specified.")
        if materialize and not loader:
            raise ValueError("The content loader must be specified.")

        self.reader = reader
        self.parser = parser
        self.loader = loader
        self.materialize = materialize
        self.include_unlabeled = include_unlabeled
        self.encoding = encoding

    @classmethod
    def from_export_file(
        cls,
        path: str,
        dataset_type: str,
        annotation_type: str,
        dataset_source_path: str,
        auth: Dict = None,
        include_unlabeled: bool = False,
        encoding: str = "utf-8",
        materialize: bool = False,
        format: str = None,
        categories: List[str] = None,
        includes_metadata=False,
    ) -> "RecordReader":
        """Initiates a RecordReader instance.

        Parameters
        ----------
        path: str
            Record file path.
        dataset_type: str
            Dataset type. Currently supports TEXT, IMAGE and DOCUMENT.
        annotation_type: str
            Annotation Type. Currently TEXT supports SINGLE_LABEL, MULTI_LABEL,
            ENTITY_EXTRACTION. IMAGE supports SINGLE_LABEL, MULTI_LABEL and BOUNDING_BOX.
            DOCUMENT supports SINGLE_LABEL and MULTI_LABEL.
        dataset_source_path: str
            Dataset source path.
        auth: (dict, optional). Default None
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        include_unlabeled: (bool, optional). Default to False.
            Whether to load the unlabeled records or not.
        encoding : (str, optional). Defaults to "utf-8".
            Encoding for text files. Used only to extract the content of the text dataset contents.
        materialize: (bool, optional). Defaults to False.
            Whether to materialize the content by loader.
        format: (str, optional). Defaults to None.
            Output format of annotations. Can be None, "spacy" for dataset
            Entity Extraction type or "yolo" for Object Detection type.
            When None, it outputs List[NERItem] or List[BoundingBoxItem].
            When "spacy", it outputs List[Tuple].
            When "yolo", it outputs List[List[Tuple]].
        categories: (List[str], optional). Defaults to None.
            The list of object categories in proper order for model training.
            Example: ['cat','dog','horse']
        includes_metadata: (bool, optional). Defaults to False.
            Determines whether the export file includes metadata or not.

        Returns
        -------
        RecordReader
            A RecordReader instance.
        """
        reader = ExportRecordReader(
            path=path, auth=auth, encoding="utf-8", includes_metadata=includes_metadata
        )
        parser = RecordParserFactory.parser(
            annotation_type=annotation_type,
            dataset_source_path=dataset_source_path,
            format=format,
            categories=categories,
        )
        loader = FileLoaderFactory.loader(
            dataset_type=dataset_type,
            auth=auth,
        )
        return cls(
            reader=reader,
            parser=parser,
            loader=loader,
            materialize=materialize,
            include_unlabeled=include_unlabeled,
            encoding=encoding,
        )

    @classmethod
    def from_DLS(
        cls,
        dataset_id: str,
        dataset_type: str,
        annotation_type: str,
        dataset_source_path: str,
        compartment_id: str = None,
        auth: Dict = None,
        include_unlabeled: bool = False,
        encoding: str = "utf-8",
        materialize: bool = False,
        format: str = None,
        categories: List[str] = None,
    ) -> "RecordReader":
        """Constructs Record Reader instance.

        Parameters
        ----------
        dataset_id: str
            The dataset OCID.
        dataset_type: str
            Dataset type. Currently supports TEXT, IMAGE and DOCUMENT.
        annotation_type: str
            Annotation Type. Currently TEXT supports SINGLE_LABEL, MULTI_LABEL,
            ENTITY_EXTRACTION. IMAGE supports SINGLE_LABEL, MULTI_LABEL and BOUNDING_BOX.
            DOCUMENT supports SINGLE_LABEL and MULTI_LABEL.
        dataset_source_path: str
            Dataset source path.
        compartment_id: (str, optional). Defaults to None.
            The compartment OCID of the dataset.
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        encoding: (str, optional). Defaults to 'utf-8'.
            Encoding for files.
        materialize: (bool, optional). Defaults to False.
            Whether the content of the dataset file should be loaded or it should return the file path to the content.
            By default the content will not be loaded.
        format: (str, optional). Defaults to None.
            Output format of annotations. Can be None, "spacy" for dataset
            Entity Extraction type or "yolo" for Object Detection type.
            When None, it outputs List[NERItem] or List[BoundingBoxItem].
            When "spacy", it outputs List[Tuple].
            When "yolo", it outputs List[List[Tuple]].
        categories: (List[str], optional). Defaults to None.
            The list of object categories in proper order for model training.
            Example: ['cat','dog','horse']

        Returns
        -------
        RecordReader
            The RecordReader instance.
        """

        if compartment_id is None:
            compartment_id = NB_SESSION_COMPARTMENT_OCID or JOB_RUN_COMPARTMENT_OCID

        if not compartment_id:
            raise ValueError("The `compartment_id` must be provided.")

        reader = DLSRecordReader(
            compartment_id=compartment_id, dataset_id=dataset_id, auth=auth
        )
        parser = DLSRecordParserFactory.parser(
            annotation_type=annotation_type,
            dataset_source_path=dataset_source_path,
            format=format,
            categories=categories,
        )
        loader = FileLoaderFactory.loader(
            dataset_type=dataset_type,
            auth=auth,
        )
        return cls(
            reader=reader,
            parser=parser,
            loader=loader,
            materialize=materialize,
            include_unlabeled=include_unlabeled,
            encoding=encoding,
        )

    def read(self) -> Generator[Tuple[str, Union[List, str]], Any, Any]:
        """Reads the record.

        Yields
        ------
        Generator[Tuple[str, Union[List, str]], Any, Any]
            File path, content and labels in a tuple.
        """
        for item in self.reader.read():
            if not item:
                return None
            record = self.parser.parse(item)
            if record.annotation or self.include_unlabeled:
                if self.materialize:
                    record.content = self.loader.load(
                        record.path, **{"encoding": self.encoding}
                    )
                yield record.to_tuple()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import click

from ads.common.auth import AuthType
from ads.opctl.conda.cmds import create as create_cmd
from ads.opctl.conda.cmds import install as install_cmd
from ads.opctl.conda.cmds import publish as publish_cmd
from ads.opctl.constants import DEFAULT_ADS_CONFIG_FOLDER
from ads.opctl.utils import suppress_traceback


@click.group("conda")
@click.help_option("--help", "-h")
def commands():
    pass


@commands.command()
@click.option("--name", "-n", help="name for the conda pack", required=False)
@click.option(
    "--version", "-v", help="version of the pack", required=False, default="1", type=str
)
@click.option(
    "--environment-file", "-f", help="path to environment file", required=True
)
@click.option(
    "--conda-pack-folder",
    help="folder to save conda pack",
    required=False,
    default=None,
)
@click.option(
    "--ads-config",
    help="folder that saves ads opctl config",
    required=False,
    default=DEFAULT_ADS_CONFIG_FOLDER,
)
@click.option(
    "--gpu",
    "-g",
    help="whether to create conda pack with gpu image",
    is_flag=True,
    default=False,
    required=False,
)
@click.option(
    "--overwrite",
    "-o",
    help="whether to overwrite downloaded pack",
    is_flag=True,
    default=False,
)
@click.option(
    "--verbose",
    is_flag=True,
    default=False,
    help="Set logging to debug for more information",
)
@click.help_option("--help", "-h")
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
def create(**kwargs):
    suppress_traceback(kwargs["debug"])(create_cmd)(**kwargs)


@commands.command()
@click.option(
    "--conda-uri", "-u", help="uri to conda pack", required=False, default=None
)
@click.option(
    "--slug", "-s", help="slug for the conda pack", required=False, default=None
)
@click.option(
    "--conda-pack-os-prefix",
    "-p",
    help="object storage prefix to download conda packs",
    required=False,
    default=None,
)
@click.option(
    "--conda-pack-folder",
    help="folder to save conda pack",
    required=False,
    default=None,
)
@click.option(
    "--oci-config",
    help="oci config file",
    required=False,
    default=None,
)
@click.option("--oci-profile", help="oci config profile", required=False, default=None)
@click.option(
    "--ads-config",
    help="folder that saves ads opctl config",
    required=False,
    default=DEFAULT_ADS_CONFIG_FOLDER,
)
@click.option(
    "--overwrite",
    "-o",
    help="whether to overwrite downloaded pack",
    is_flag=True,
    default=False,
)
@click.help_option("--help", "-h")
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
@click.option(
    "--auth",
    "-a",
    help="authentication method",
    type=click.Choice(AuthType.values()),
    default=None,
)
def install(**kwargs):
    suppress_traceback(kwargs["debug"])(install_cmd)(**kwargs)


@commands.command()
@click.option("--slug", "-s", help="slug for the conda pack", required=False)
@click.option("--name", "-n", help="name for the conda pack", required=False)
@click.option(
    "--version",
    "-v",
    help="version of the pack",
    required=False,
    default="1",
    type=str,
)
@click.option(
    "--environment-file",
    "-f",
    help="conda environment yaml path",
    required=False,
    default=None,
)
@click.option(
    "--conda-pack-os-prefix",
    "-p",
    help="object storage prefix to save published conda packs, in the form oci://<bucket>@<namespace>/path/",
    required=False,
    default=None,
)
@click.option(
    "--conda-pack-folder",
    help="folder to save conda pack",
    required=False,
    default=None,
)
@click.option(
    "--gpu",
    "-g",
    help="whether to create conda pack with gpu image",
    is_flag=True,
    default=False,
    required=False,
)
@click.option(
    "--oci-config",
    help="oci config file",
    required=False,
    default=None,
)
@click.option("--oci-profile", help="oci config profile", required=False, default=None)
@click.option(
    "--ads-config",
    help="folder that saves ads opctl config",
    required=False,
    default=DEFAULT_ADS_CONFIG_FOLDER,
)
@click.option(
    "--overwrite",
    "-o",
    help="whether to overwrite downloaded pack",
    is_flag=True,
    default=False,
)
@click.help_option("--help", "-h")
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
@click.option(
    "--auth",
    "-a",
    help="authentication method",
    type=click.Choice(AuthType.values()),
    default=None,
)
def publish(**kwargs):
    suppress_traceback(kwargs["debug"])(publish_cmd)(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import configparser
import os
from typing import List

import urllib.parse
import fsspec
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


class OperatorNotFound(Exception):
    pass


class CondaPackInfoNotProvided(Exception):
    pass


class NotSupportedError(Exception):
    pass


class _DefaultNoneDict(dict):
    def __missing__(self, key):
        return ""


def read_from_ini(path: str) -> configparser.ConfigParser:
    if not os.path.exists(path):
        raise FileNotFoundError(f"File {path} is not found.")
    parser = configparser.ConfigParser(default_section=None)
    parser.optionxform = str  # preserve case
    parser.read(path)
    return parser


@runtime_dependency(module="nbformat", install_from=OptionalDependency.OPCTL)
@runtime_dependency(module="nbconvert", install_from=OptionalDependency.OPCTL)
def convert_notebook(
    input_path,
    auth,
    exclude_tags: List = None,
    output_path: str = None,
    overwrite: bool = False,
) -> str:
    with fsspec.open(input_path, **auth) as f:
        nb = nbformat.reads(f.read(), nbformat.NO_CONVERT)

    from nbconvert.preprocessors import TagRemovePreprocessor
    from nbconvert.exporters import PythonExporter

    exporter = PythonExporter()
    if exclude_tags:
        exporter.register_preprocessor(
            TagRemovePreprocessor(remove_cell_tags=exclude_tags), enabled=True
        )

    output, _ = exporter.from_notebook_node(nb)
    output = output.strip("\n")
    output_path = output_path or os.path.splitext(input_path)[0] + ".py"
    file_system_clz = fsspec.get_filesystem_class(
        urllib.parse.urlparse(output_path).scheme or "file"
    )
    file_system = file_system_clz(**auth)
    if not overwrite and file_system.exists(output_path):
        raise FileExistsError(
            f"{output_path} exists. Please rename your notebook or use overwrite option."
        )
    with fsspec.open(output_path, mode="wb", **auth) as f:
        f.write(output.encode("utf-8"))
    return output_path



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging
from ads.feature_engineering.feature_type.handler import warnings as feature_warning
from ads.feature_engineering.feature_type_manager import (
    FeatureTypeManager as feature_type_manager,
)
from ads.feature_engineering.feature_type.base import FeatureType
from ads.feature_engineering.feature_type.base import Tag
from ads.feature_engineering.adsimage.image import ADSImage
from ads.feature_engineering.adsimage.image_reader import ADSImageReader
from ads.feature_engineering.feature_type.adsstring.oci_language import OCILanguage

feature_type = feature_type_manager.feature_type_registered

logger = logging.getLogger(__name__)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import fsspec
import io
import json
import os
from typing import Dict, Generator, List, Union

from ads.common.decorator.runtime_dependency import runtime_dependency
from ads.text_dataset.utils import PY4JGateway, experimental
from fsspec.core import OpenFile


class Base:
    """Base class for backends."""

    def read_line(
        self, fhandler: OpenFile
    ) -> Generator[Union[str, List[str]], None, None]:
        """Read lines from a file.

        Parameters
        ----------
        fhandler : `fsspec.core.OpenFile`
            a file handler returned by `fsspec`

        Yields
        -------
        Generator
            a generator that yields lines
        """
        with fhandler as f:
            for line in f:
                yield line.decode(fhandler.encoding)

    def read_text(
        self, fhandler: OpenFile
    ) -> Generator[Union[str, List[str]], None, None]:
        """Read entire file into a string.

        Parameters
        ----------
        fhandler : `fsspec.core.OpenFile`
            a file handler returned by `fsspec`

        Yields
        -------
        Generator
            a generator that yields text in the file
        """
        with fhandler as f:
            yield f.read().decode(fhandler.encoding)

    @staticmethod
    def _validate_dest(src, dst_folder: str, fname: str = None) -> str:
        if fname is None:
            fname = os.path.splitext(os.path.basename(src))[0] + ".txt"
        return os.path.join(dst_folder, fname)

    def convert_to_text(
        self,
        fhandler: OpenFile,
        dst_path: str,
        fname: str = None,
        storage_options: Dict = None,
    ) -> str:
        """Convert input file to a text file

        Parameters
        ----------
        fhandler : `fsspec.core.OpenFile`
            a file handler returned by `fsspec`
        dst_path: str
            local folder or cloud storage prefix to save converted text files
        fname: str, optional
            filename for converted output, relative to dirname or prefix, by default None
        storage_options: dict, optional
            storage options for cloud storage

        Returns
        -------
        str
            path to saved output
        """
        dest = self._validate_dest(fhandler.path, dst_path, fname)
        storage_options = {} if storage_options is None else storage_options
        with fsspec.open(dest, mode="wb", **storage_options) as fout:
            with fhandler as fin:
                fout.write(fin.read())
        return dest

    def get_metadata(self, fhandler: OpenFile) -> Dict:
        """Get metadata of a file.

        Parameters
        ----------
        fhandler : `fsspec.core.OpenFile`
            a file handler returned by fsspec

        Returns
        -------
        dict
            dictionary of metadata
        """
        return {}


class Tika(Base):
    def read_line(self, fhandler):
        with PY4JGateway() as gateway:
            parser = gateway.jvm.parsers.TikaParser()
            with fhandler as f:
                reader = parser.reader(f.read())
            while True:
                line = reader.readLine()
                if line is not None:
                    yield line
                else:
                    break

    def read_text(self, fhandler):
        with PY4JGateway() as gateway:
            parser = gateway.jvm.parsers.TikaParser()
            with fhandler as f:
                yield parser.readText(f.read()).decode()

    def convert_to_text(self, fhandler, dst_path, fname=None, storage_options=None):
        dest = self._validate_dest(fhandler.path, dst_path, fname)
        storage_options = {} if storage_options is None else storage_options
        with fsspec.open(dest, mode="w", **storage_options) as f:
            f.write(next(self.read_text(fhandler)))
        return dest

    def get_metadata(self, fhandler):
        with PY4JGateway() as gateway:
            parser = gateway.jvm.parsers.TikaParser()
            meta = gateway.jvm.org.apache.tika.metadata.Metadata()
            with fhandler as f:
                parser.readText(f.read(), meta)
            return json.loads(parser.getMetadata(meta))

    def detect_encoding(self, fhandler: OpenFile):
        with PY4JGateway() as gateway:
            with fhandler as f:
                return gateway.jvm.parsers.TikaParser.detectEncoding(f.read())


class PDFPlumber(Base):
    @runtime_dependency(
        module="pdfplumber", err_msg="pdfplumber must be installed first."
    )
    def __init__(self):
        super().__init__()

    def read_line(self, fhandler):
        import pdfplumber

        with fhandler as f:
            pdf = pdfplumber.PDF(f)
            for page in pdf.pages:
                reader = io.StringIO(page.extract_text())
                for line in reader:
                    yield line

    def read_text(self, fhandler):
        import pdfplumber

        with fhandler as f:
            pdf = pdfplumber.PDF(f)
            texts = (page.extract_text() for page in pdf.pages)
            yield "\n".join([text for text in texts if text is not None])

    def convert_to_text(
        self,
        fhandler,
        dst_path,
        fname=None,
        storage_options=None,
    ):
        import pdfplumber

        dest = self._validate_dest(fhandler.path, dst_path, fname)
        storage_options = {} if storage_options is None else storage_options
        with fhandler as fin:
            pdf = pdfplumber.PDF(fin)
            with fsspec.open(dest, mode="w", **storage_options) as fout:
                for page in pdf.pages:
                    text = page.extract_text()
                    if text is not None:
                        fout.write(text)
        return dest

    def get_metadata(self, fhandler):
        import pdfplumber

        with fhandler as f:
            pdf = pdfplumber.PDF(f)
            return pdf.metadata


@experimental
class OITCC(Base):  # pragma: no cover
    def convert_to_text(self, fhandler, output_path, fname=None, storage_options=None):
        dest = self._validate_dest(fhandler.path, output_path, fname)
        storage_options = {} if storage_options is None else storage_options
        with fsspec.open(dest, mode="w", **storage_options) as f:
            f.write(next(self.read_text(fhandler)))



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
from typing import Union
from ads.jobs.builders.runtimes.base import Runtime


class ContainerRuntime(Runtime):
    """Represents a container job runtime

    To define container runtime:

    >>> ContainerRuntime()
    >>> .with_image("iad.ocir.io/<your_tenancy>/<your_image>")
    >>> .with_cmd("sleep 5 && echo Hello World")
    >>> .with_entrypoint(["/bin/sh", "-c"])
    >>> .with_environment_variable(MY_ENV="MY_VALUE")

    Alternatively, you can define the ``entrypoint`` and ``cmd`` along with the image.

    >>> ContainerRuntime()
    >>> .with_image(
    >>>     "iad.ocir.io/<your_tenancy>/<your_image>",
    >>>     entrypoint=["/bin/sh", -c],
    >>>     cmd="sleep 5 && echo Hello World",
    >>> )
    >>> .with_environment_variable(MY_ENV="MY_VALUE")

    The entrypoint and cmd can be either "exec form" or "shell form" (See references).
    The exec form is used when a list is passed in.
    The shell form is used when a space separated string is passed in.

    When using the ContainerRuntime with OCI Data Science Job, the exec form is recommended.
    For most images, when the entrypoint is set to ``["/bin/sh", "-c"]``,
    ``cmd`` can be a string as if you are running shell command.

    References
    ----------
    https://docs.docker.com/engine/reference/builder/#entrypoint
    https://docs.docker.com/engine/reference/builder/#cmd

    """

    CONST_IMAGE = "image"
    CONST_ENTRYPOINT = "entrypoint"
    CONST_CMD = "cmd"
    attribute_map = {
        CONST_IMAGE: CONST_IMAGE,
        CONST_ENTRYPOINT: CONST_ENTRYPOINT,
        CONST_CMD: CONST_CMD,
    }
    attribute_map.update(Runtime.attribute_map)

    @property
    def image(self) -> str:
        """The container image"""
        return self.get_spec(self.CONST_IMAGE)

    def with_image(
        self, image: str, entrypoint: Union[str, list, None] = None, cmd: str = None
    ) -> "ContainerRuntime":
        """Specify the image for the container job.

        Parameters
        ----------
        image : str
            The container image, e.g. iad.ocir.io/<your_tenancy>/<your_image>:<your_tag>
        entrypoint : str or list, optional
            Entrypoint for the job, by default None (the entrypoint defined in the image will be used).
        cmd : str, optional
            Command for the job, by default None.

        Returns
        -------
        ContainerRuntime
            The runtime instance.
        """
        self.with_entrypoint(entrypoint)
        self.set_spec(self.CONST_CMD, cmd)
        return self.set_spec(self.CONST_IMAGE, image)

    @property
    def entrypoint(self) -> str:
        """Entrypoint of the container job"""
        return self.get_spec(self.CONST_ENTRYPOINT)

    def with_entrypoint(self, entrypoint: Union[str, list]) -> "ContainerRuntime":
        """Specifies the entrypoint for the container job.

        Parameters
        ----------
        entrypoint : str or list
            Entrypoint for the container job

        Returns
        -------
        ContainerRuntime
            The runtime instance.
        """
        self._spec[self.CONST_ENTRYPOINT] = entrypoint
        return self

    @property
    def cmd(self) -> str:
        """Command of the container job"""
        return self.get_spec(self.CONST_CMD)

    def with_cmd(self, cmd: str) -> "ContainerRuntime":
        """Specifies the command for the container job.

        Parameters
        ----------
        cmd : str
            Command for the container job

        Returns
        -------
        ContainerRuntime
            The runtime instance.
        """
        self._spec[self.CONST_CMD] = cmd
        return self



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import cloudpickle
import numpy as np
import pandas as pd
from ads.model.serde.common import Serializer, Deserializer
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common import logger
from pandas.api.types import is_numeric_dtype, is_string_dtype
from typing import Any, Dict, List, Optional, Tuple, Union
from joblib import dump


MODEL_SERIALIZATION_TYPE_ONNX = "onnx"
MODEL_SERIALIZATION_TYPE_CLOUDPICKLE = "cloudpickle"
MODEL_SERIALIZATION_TYPE_TORHCSCRIPT = "torchscript"
MODEL_SERIALIZATION_TYPE_TORCH = "torch"
MODEL_SERIALIZATION_TYPE_TORCH_ONNX = "torch_onnx"
MODEL_SERIALIZATION_TYPE_TF = "tf"
MODEL_SERIALIZATION_TYPE_TF_ONNX = "tf_onnx"
MODEL_SERIALIZATION_TYPE_JOBLIB = "joblib"
MODEL_SERIALIZATION_TYPE_SKLEARN_ONNX = "sklearn_onnx"
MODEL_SERIALIZATION_TYPE_LIGHTGBM = "lightgbm"
MODEL_SERIALIZATION_TYPE_LIGHTGBM_ONNX = "lightgbm_onnx"
MODEL_SERIALIZATION_TYPE_XGBOOST = "xgboost"
MODEL_SERIALIZATION_TYPE_XGBOOST_UBJ = "xgboost_ubj"
MODEL_SERIALIZATION_TYPE_XGBOOST_TXT = "xgboost_txt"
MODEL_SERIALIZATION_TYPE_XGBOOST_ONNX = "xgboost_onnx"
MODEL_SERIALIZATION_TYPE_SPARK = "spark"
MODEL_SERIALIZATION_TYPE_HUGGINGFACE = "huggingface"


SUPPORTED_MODEL_SERIALIZERS = [
    MODEL_SERIALIZATION_TYPE_ONNX,
    MODEL_SERIALIZATION_TYPE_CLOUDPICKLE,
    MODEL_SERIALIZATION_TYPE_TORHCSCRIPT,
    MODEL_SERIALIZATION_TYPE_TORCH,
    MODEL_SERIALIZATION_TYPE_TORCH_ONNX,
    MODEL_SERIALIZATION_TYPE_TF,
    MODEL_SERIALIZATION_TYPE_TF_ONNX,
    MODEL_SERIALIZATION_TYPE_JOBLIB,
    MODEL_SERIALIZATION_TYPE_SKLEARN_ONNX,
    MODEL_SERIALIZATION_TYPE_LIGHTGBM,
    MODEL_SERIALIZATION_TYPE_LIGHTGBM_ONNX,
    MODEL_SERIALIZATION_TYPE_XGBOOST,
    MODEL_SERIALIZATION_TYPE_XGBOOST_ONNX,
    MODEL_SERIALIZATION_TYPE_SPARK,
    MODEL_SERIALIZATION_TYPE_HUGGINGFACE,
]


class ModelSerializerType:
    CLOUDPICKLE = MODEL_SERIALIZATION_TYPE_CLOUDPICKLE
    ONNX = MODEL_SERIALIZATION_TYPE_ONNX


class PyTorchModelSerializerType:
    TORCH = MODEL_SERIALIZATION_TYPE_TORCH
    TORCHSCRIPT = MODEL_SERIALIZATION_TYPE_TORHCSCRIPT
    ONNX = MODEL_SERIALIZATION_TYPE_TORCH_ONNX


class TensorflowModelSerializerType:
    TENSORFLOW = MODEL_SERIALIZATION_TYPE_TF
    ONNX = MODEL_SERIALIZATION_TYPE_TF_ONNX


class LightGBMModelSerializerType:
    LIGHTGBM = MODEL_SERIALIZATION_TYPE_LIGHTGBM
    ONNX = MODEL_SERIALIZATION_TYPE_LIGHTGBM_ONNX


class SklearnModelSerializerType:
    JOBLIB = MODEL_SERIALIZATION_TYPE_JOBLIB
    CLOUDPICKLE = MODEL_SERIALIZATION_TYPE_CLOUDPICKLE
    ONNX = MODEL_SERIALIZATION_TYPE_SKLEARN_ONNX


class XgboostModelSerializerType:
    XGBOOST = MODEL_SERIALIZATION_TYPE_XGBOOST
    ONNX = MODEL_SERIALIZATION_TYPE_XGBOOST_ONNX


class SparkModelSerializerType:
    SPARK = MODEL_SERIALIZATION_TYPE_SPARK


class HuggingFaceSerializerType:
    HUGGINGFACE = MODEL_SERIALIZATION_TYPE_HUGGINGFACE


class ModelSerializer(Serializer):
    """Base class for creation of new model serializers."""

    def __init__(self, model_file_suffix):
        super().__init__()
        self.model_file_suffix = model_file_suffix


class ModelDeserializer(Deserializer):
    """Base class for creation of new model deserializers."""

    def deserialize(self, **kwargs):
        raise NotImplementedError


class CloudPickleModelSerializer(ModelSerializer):
    """Uses `Cloudpickle` to save model."""

    def __init__(self, model_file_suffix="pkl"):
        super().__init__(model_file_suffix=model_file_suffix)

    def serialize(self, estimator, model_path, **kwargs):
        """Uses `cloudpickle.dump` to save model. See https://docs.python.org/3/library/pickle.html#pickle.dump for more details.

        Args:
            estimator: The model to be saved.
            model_path: The file object or path of the model in which it is to be stored.
            kwargs:
                model_save: (dict, optional).
                    The dictionary where contains the availiable options to be passed to `cloudpickle.dump`.
        """
        cloudpickle_kwargs = kwargs.pop("model_save", {})
        with open(model_path, "wb") as f:
            cloudpickle.dump(estimator, f, **cloudpickle_kwargs)


class JobLibModelSerializer(ModelSerializer):
    """Uses `Joblib` to save model."""

    def __init__(self, model_file_suffix="joblib"):
        super().__init__(model_file_suffix=model_file_suffix)

    def serialize(self, estimator, model_path, **kwargs):
        """Uses `joblib.dump` to save model. See https://joblib.readthedocs.io/en/latest/generated/joblib.dump.html for more details.

        Args:
            estimator: The model to be saved.
            model_path: The file object or path of the model in which it is to be stored.
            kwargs:
                model_save: (dict, optional).
                    The dictionary where contains the availiable options to be passed to `joblib.dump`.
        """
        joblib_kwargs = kwargs.pop("model_save", {})
        dump(estimator, model_path, **joblib_kwargs)


class SparkModelSerializer(ModelSerializer):
    """Save Spark Model."""

    def __init__(self, model_file_suffix=""):
        super().__init__(model_file_suffix=model_file_suffix)

    def serialize(self, estimator, model_path, **kwargs):
        estimator.write().overwrite().save(model_path)


class PyTorchModelSerializer(ModelSerializer):
    """Save PyTorch Model using torch.save(). See https://pytorch.org/docs/stable/generated/torch.save.html for more details."""

    def __init__(self, model_file_suffix="pt"):
        super().__init__(model_file_suffix=model_file_suffix)

    @runtime_dependency(module="torch", install_from=OptionalDependency.PYTORCH)
    def serialize(self, estimator, model_path, **kwarg):
        torch.save(estimator.state_dict(), model_path)


class TorchScriptModelSerializer(ModelSerializer):
    """Save PyTorch Model using torchscript. See https://pytorch.org/tutorials/beginner/saving_loading_models.html#export-load-model-in-torchscript-format for more details."""

    def __init__(self, model_file_suffix="pt"):
        super().__init__(model_file_suffix=model_file_suffix)

    @runtime_dependency(module="torch", install_from=OptionalDependency.PYTORCH)
    def serialize(self, estimator, model_path, **kwargs):
        compiled_model = torch.jit.script(estimator)
        torch.jit.save(compiled_model, model_path)


class LightGBMModelSerializer(ModelSerializer):
    """Save LightGBM Model through save_model into txt."""

    def __init__(self, model_file_suffix="txt"):
        super().__init__(model_file_suffix=model_file_suffix)

    def serialize(self, estimator, model_path, **kwargs):
        estimator.save_model(model_path)


class XgboostJsonModelSerializer(ModelSerializer):
    """Save Xgboost Model through xgboost.save_model into JSON."""

    def __init__(self, model_file_suffix="json"):
        super().__init__(model_file_suffix=model_file_suffix)

    def serialize(self, estimator, model_path, **kwargs):
        """Save Xgboost Model through xgboost.save_model .See
        https://xgboost.readthedocs.io/en/stable/python/python_api.html#xgboost.Booster.save_model
        for more details.

        Args:
            estimator: The model to be saved.
            model_path: The file object or path of the model in which it is to be stored.
        """
        estimator.save_model(model_path)


class XgboostTxtModelSerializer(ModelSerializer):
    """Save Xgboost Model through xgboost.save_model into txt."""

    def __init__(self, model_file_suffix="txt"):
        super().__init__(model_file_suffix=model_file_suffix)

    def serialize(self, estimator, model_path, **kwargs):
        """Save Xgboost Model through xgboost.save_model .See
        https://xgboost.readthedocs.io/en/stable/python/python_api.html#xgboost.Booster.save_model
        for more details.

        Args:
            estimator: The model to be saved.
            model_path: The file object or path of the model in which it is to be stored.
        """
        estimator.save_model(model_path)


class XgboostUbjModelSerializer(ModelSerializer):
    """Save Xgboost Model through xgboost.save_model into binary JSON."""

    def __init__(self, model_file_suffix="ubj"):
        super().__init__(model_file_suffix=model_file_suffix)

    def serialize(self, estimator, model_path, **kwargs):
        """Save Xgboost Model through xgboost.save_model .See
        https://xgboost.readthedocs.io/en/stable/python/python_api.html#xgboost.Booster.save_model
        for more details.

        Args:
            estimator: The model to be saved.
            model_path: The file object or path of the model in which it is to be stored.
        """
        estimator.save_model(model_path)


class TensorFlowModelSerializer(ModelSerializer):
    """Save Tensorflow Model."""

    def __init__(self, model_file_suffix="h5"):
        super().__init__(model_file_suffix=model_file_suffix)

    def serialize(self, estimator, model_path, **kwargs):
        estimator.save(model_path)


class HuggingFaceModelSerializer(ModelSerializer):
    """Save HuggingFace Pipeline."""

    def __init__(self, model_file_suffix=""):
        super().__init__(model_file_suffix=model_file_suffix)

    def serialize(self, estimator, model_path, **kwargs):
        estimator.save_pretrained(save_directory=model_path)
        estimator.model.config.use_pretrained_backbone = False
        estimator.model.config.save_pretrained(save_directory=model_path)


class OnnxModelSerializer(ModelSerializer):
    """Base class for creation of onnx converter for each model framework."""

    def __init__(self, model_file_suffix="onnx"):
        super().__init__(model_file_suffix=model_file_suffix)

    def serialize(
        self,
        estimator,
        model_path,
        initial_types: List[Tuple] = None,
        X_sample: Optional[
            Union[
                Dict,
                str,
                List,
                Tuple,
                np.ndarray,
                pd.core.series.Series,
                pd.core.frame.DataFrame,
            ]
        ] = None,
        **kwargs,
    ):
        """Save model into onnx format.

        Args:
            estimator: The model to be saved.
            model_path: The file object or path of the model in which it is to be stored.
            initial_types: (List[Tuple], optional)
                a python list. Each element is a tuple of a variable name and a data type.
            X_sample: (any, optional). Defaults to None.
                Contains model inputs such that model(X_sample) is a valid
                invocation of the model, used to valid model input type.
        """
        self.estimator = estimator
        onx = self._to_onnx(
            initial_types=initial_types,
            X_sample=X_sample,
            **kwargs,
        )
        with open(model_path, "wb") as f:
            f.write(onx.SerializeToString())

    def _to_onnx(
        self,
        initial_types: List[Tuple] = None,
        X_sample: Optional[
            Union[
                Dict,
                str,
                List,
                Tuple,
                np.ndarray,
                pd.core.series.Series,
                pd.core.frame.DataFrame,
            ]
        ] = None,
        **kwargs,
    ):
        raise NotImplementedError


class SklearnOnnxModelSerializer(OnnxModelSerializer):
    """Converts Skearn Model into Onnx."""

    def __init__(self):
        super().__init__()

    @runtime_dependency(module="onnx", install_from=OptionalDependency.ONNX)
    @runtime_dependency(module="xgboost", install_from=OptionalDependency.BOOSTED)
    @runtime_dependency(module="lightgbm", install_from=OptionalDependency.BOOSTED)
    @runtime_dependency(module="skl2onnx", install_from=OptionalDependency.ONNX)
    @runtime_dependency(module="onnxmltools", install_from=OptionalDependency.ONNX)
    @runtime_dependency(
        module="onnxmltools.convert.xgboost.operator_converters.XGBoost",
        object="convert_xgboost",
        install_from=OptionalDependency.ONNX,
    )
    @runtime_dependency(
        module="onnxmltools.convert.lightgbm.operator_converters.LightGbm",
        object="convert_lightgbm",
        install_from=OptionalDependency.ONNX,
    )
    def _to_onnx(
        self,
        initial_types: List[Tuple] = None,
        X_sample: Optional[
            Union[
                Dict,
                str,
                List,
                Tuple,
                np.ndarray,
                pd.core.series.Series,
                pd.core.frame.DataFrame,
            ]
        ] = None,
        **kwargs,
    ):
        """
        Produces an equivalent ONNX model of the given scikit-learn model.

        Parameters
        ----------
        initial_types: (List[Tuple], optional). Defaults to None.
            Each element is a tuple of a variable name and a type.
        X_sample: Union[Dict, str, List, np.ndarray, pd.core.series.Series, pd.core.frame.DataFrame,]. Defaults to None.
            Contains model inputs such that model(X_sample) is a valid invocation of the model.
            Used to generate initial_types.

        Returns
        -------
        onnx.onnx_ml_pb2.ModelProto
            An ONNX model (type: ModelProto) which is equivalent to the input scikit-learn model.
        """
        auto_generated_initial_types = None
        if not initial_types:
            if X_sample is None:
                raise ValueError(
                    " At least one of `X_sample` or `initial_types` must be provided."
                )
            auto_generated_initial_types = self._generate_initial_types(X_sample)
        if str(type(self.estimator)).startswith("<class 'sklearn.pipeline"):
            model_types = []
            model_types = [type(val[1]) for val in self.estimator.steps]
            if xgboost.sklearn.XGBClassifier in model_types:
                skl2onnx.update_registered_converter(
                    xgboost.XGBClassifier,
                    "XGBoostXGBClassifier",
                    skl2onnx.common.shape_calculator.calculate_linear_classifier_output_shapes,
                    convert_xgboost,
                    options=kwargs.pop(
                        "options", {"nocl": [True, False], "zipmap": [True, False]}
                    ),
                )

            if xgboost.sklearn.XGBRegressor in model_types:
                skl2onnx.update_registered_converter(
                    xgboost.XGBRegressor,
                    "XGBoostXGBRegressor",
                    skl2onnx.common.shape_calculator.calculate_linear_regressor_output_shapes,
                    convert_xgboost,
                )

            if lightgbm.sklearn.LGBMClassifier in model_types:
                skl2onnx.update_registered_converter(
                    lightgbm.LGBMClassifier,
                    "LightGbmLGBMClassifier",
                    skl2onnx.common.shape_calculator.calculate_linear_classifier_output_shapes,
                    convert_lightgbm,
                    options=kwargs.pop(
                        "options",
                        {"nocl": [True, False], "zipmap": [True, False, "columns"]},
                    ),
                )

            if lightgbm.sklearn.LGBMRegressor in model_types:

                def skl2onnx_convert_lightgbm(scope, operator, container):
                    options = scope.get_options(operator.raw_operator)
                    if "split" in options:
                        if StrictVersion(onnxmltools.__version__) < StrictVersion(
                            "1.9.2"
                        ):
                            logger.warnings(
                                "Option split was released in version 1.9.2 but %s is "
                                "installed. It will be ignored."
                                % onnxmltools.__version__
                            )
                        operator.split = options["split"]
                    else:
                        operator.split = None
                    convert_lightgbm(scope, operator, container)

                skl2onnx.update_registered_converter(
                    lightgbm.LGBMRegressor,
                    "LightGbmLGBMRegressor",
                    skl2onnx.common.shape_calculator.calculate_linear_regressor_output_shapes,
                    skl2onnx_convert_lightgbm,
                    options=kwargs.pop("options", {"split": None}),
                )
            if initial_types:
                return skl2onnx.convert_sklearn(
                    self.estimator, initial_types=initial_types, **kwargs
                )
            else:
                try:
                    return skl2onnx.convert_sklearn(
                        self.estimator,
                        initial_types=auto_generated_initial_types,
                        target_opset=None,
                        **kwargs,
                    )
                except Exception as e:
                    raise ValueError(
                        "`initial_types` can not be autodetected. Please directly pass `initial_types`."
                    )
        else:
            if initial_types:
                return onnxmltools.convert_sklearn(
                    self.estimator,
                    initial_types=initial_types,
                    targeted_onnx=onnx.__version__,
                    **kwargs,
                )
            else:
                try:
                    return onnxmltools.convert_sklearn(
                        self.estimator,
                        initial_types=auto_generated_initial_types,
                        targeted_onnx=onnx.__version__,
                        **kwargs,
                    )
                except Exception as e:
                    raise ValueError(
                        "`initial_types` can not be detected. Please directly pass initial_types."
                    )

    @runtime_dependency(module="skl2onnx", install_from=OptionalDependency.ONNX)
    def _generate_initial_types(self, X_sample: Any) -> List:
        """Auto generate intial types.

        Parameters
        ----------
        X_sample: (Any)
            Train data.

        Returns
        -------
        List
            Initial types.
        """
        if self._is_all_numerical_array_dataframe(X_sample):
            # if it's a dataframe and all the columns are numerical. Or
            # it's not a dataframe, also try this.
            if hasattr(X_sample, "shape") and len(X_sample.shape) >= 2:
                auto_generated_initial_types = [
                    (
                        "input",
                        skl2onnx.common.data_types.FloatTensorType(
                            [None, X_sample.shape[1]]
                        ),
                    )
                ]
            elif hasattr(self.estimator, "n_features_in_"):
                n_cols = self.estimator.n_features_in_
                auto_generated_initial_types = [
                    (
                        "input",
                        skl2onnx.common.data_types.FloatTensorType([None, n_cols]),
                    )
                ]
            else:
                raise ValueError(
                    "`initial_types` can not be detected. Please directly pass initial_types."
                )
        elif self.is_either_numerical_or_string_dataframe(X_sample):
            # for dataframe and not all the columns are numerical, then generate
            # the input types of all the columns one by one.
            auto_generated_initial_types = []

            for i, col in X_sample.iteritems():
                if is_numeric_dtype(col.dtypes):
                    auto_generated_initial_types.append(
                        (
                            col.name,
                            skl2onnx.common.data_types.FloatTensorType([None, 1]),
                        )
                    )
                else:
                    auto_generated_initial_types.append(
                        (
                            col.name,
                            skl2onnx.common.data_types.StringTensorType([None, 1]),
                        )
                    )
        else:
            try:
                auto_generated_initial_types = (
                    skl2onnx.common.data_types.guess_data_type(
                        np.array(X_sample) if isinstance(X_sample, list) else X_sample
                    )
                )
            except:
                auto_generated_initial_types = None
        return auto_generated_initial_types

    @staticmethod
    def _is_all_numerical_array_dataframe(
        data: Union[pd.DataFrame, np.ndarray]
    ) -> bool:
        """Check whether all the columns are numerical for numpy array and dataframe.
        For data with any other data types, it will return False.

        Parameters
        ----------
        data: Union[pd.DataFrame, np.ndarray]

        Returns
        -------
        bool
            Whether all the columns in a pandas dataframe or numpy array are all numerical.
        """
        return (
            isinstance(data, pd.DataFrame)
            and all([is_numeric_dtype(dtype) for dtype in data.dtypes])
            or (isinstance(data, np.ndarray) and is_numeric_dtype(data.dtype))
        )

    @staticmethod
    def is_either_numerical_or_string_dataframe(data: pd.DataFrame) -> bool:
        """Check whether all the columns are either numerical or string for dataframe."""
        return isinstance(data, pd.DataFrame) and all(
            [
                is_numeric_dtype(col.dtypes) or is_string_dtype(col.dtypes)
                for _, col in data.iteritems()
            ]
        )


class LightGBMOnnxModelSerializer(OnnxModelSerializer):
    """Converts LightGBM model into onnx format."""

    def __init__(self):
        super().__init__()

    @runtime_dependency(
        module="skl2onnx.common.data_types",
        object="FloatTensorType",
        install_from=OptionalDependency.ONNX,
    )
    @runtime_dependency(
        module="onnxmltools.convert",
        object="convert_lightgbm",
        install_from=OptionalDependency.ONNX,
    )
    def _to_onnx(
        self,
        initial_types: List[Tuple] = None,
        X_sample: Optional[
            Union[
                Dict,
                str,
                List,
                Tuple,
                np.ndarray,
                pd.core.series.Series,
                pd.core.frame.DataFrame,
            ]
        ] = None,
        **kwargs,
    ):
        """
        Produces an equivalent ONNX model of the given LightGBM model.

        Parameters
        ----------
        initial_types: (List[Tuple], optional). Defaults to None.
            Each element is a tuple of a variable name and a type.
        X_sample: Union[Dict, str, List, np.ndarray, pd.core.series.Series, pd.core.frame.DataFrame,]. Defaults to None.
            Contains model inputs such that model(X_sample) is a valid invocation of the model.
            Used to generate initial_types.

        Returns
        ------
            An ONNX model (type: ModelProto) which is equivalent to the input LightGBM model.
        """
        auto_generated_initial_types = None
        if not initial_types:
            auto_generated_initial_types = self._generate_initial_types(X_sample)
            try:
                return convert_lightgbm(
                    self.estimator,
                    initial_types=auto_generated_initial_types,
                    target_opset=kwargs.pop("target_opset", None),
                    **kwargs,
                )
            except:
                raise ValueError(
                    "`initial_types` can not be detected. Please directly pass initial_types."
                )
        else:
            return convert_lightgbm(
                self.estimator,
                initial_types=initial_types,
                target_opset=kwargs.pop("target_opset", None),
                **kwargs,
            )

    @runtime_dependency(
        module="skl2onnx.common.data_types",
        object="FloatTensorType",
        install_from=OptionalDependency.ONNX,
    )
    def _generate_initial_types(self, X_sample: Any) -> List:
        """Auto generate intial types.

        Parameters
        ----------
        X_sample: (Any)
            Train data.

        Returns
        -------
        List
            Initial types.
        """
        if X_sample is not None and hasattr(X_sample, "shape"):
            auto_generated_initial_types = [
                ("input", FloatTensorType([None, X_sample.shape[1]]))
            ]
        elif hasattr(self.estimator, "num_feature"):
            n_cols = self.estimator.num_feature()
            auto_generated_initial_types = [("input", FloatTensorType([None, n_cols]))]
        elif hasattr(self.estimator, "n_features_in_"):
            n_cols = self.estimator.n_features_in_
            auto_generated_initial_types = [("input", FloatTensorType([None, n_cols]))]
        else:
            raise ValueError(
                "`initial_types` can not be detected. Please directly pass initial_types."
            )
        return auto_generated_initial_types


class XgboostOnnxModelSerializer(OnnxModelSerializer):
    """Converts Xgboost model into onnx format."""

    def __init__(self):
        super().__init__()

    @runtime_dependency(module="onnx", install_from=OptionalDependency.ONNX)
    @runtime_dependency(module="xgboost", install_from=OptionalDependency.BOOSTED)
    @runtime_dependency(
        module="skl2onnx",
        object="convert_sklearn",
        install_from=OptionalDependency.ONNX,
    )
    @runtime_dependency(
        module="skl2onnx",
        object="update_registered_converter",
        install_from=OptionalDependency.ONNX,
    )
    @runtime_dependency(
        module="skl2onnx.common.data_types",
        object="FloatTensorType",
        install_from=OptionalDependency.ONNX,
    )
    @runtime_dependency(
        module="skl2onnx.common.shape_calculator",
        object="calculate_linear_classifier_output_shapes",
        install_from=OptionalDependency.ONNX,
    )
    @runtime_dependency(
        module="skl2onnx.common.shape_calculator",
        object="calculate_linear_regressor_output_shapes",
        install_from=OptionalDependency.ONNX,
    )
    @runtime_dependency(module="onnxmltools", install_from=OptionalDependency.ONNX)
    @runtime_dependency(
        module="onnxmltools.convert.xgboost.operator_converters.XGBoost",
        object="convert_xgboost",
        install_from=OptionalDependency.ONNX,
    )
    def _to_onnx(
        self,
        initial_types: List[Tuple] = None,
        X_sample: Union[list, tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        **kwargs,
    ):
        """
        Produces an equivalent ONNX model of the given Xgboost model.

        Parameters
        ----------
        initial_types: (List[Tuple], optional). Defaults to None.
            Each element is a tuple of a variable name and a type.
        X_sample: Union[Dict, str, List, np.ndarray, pd.core.series.Series, pd.core.frame.DataFrame,]. Defaults to None.
            Contains model inputs such that model(X_sample) is a valid invocation of the model.
            Used to generate initial_types.

        Returns
        -------
        onnx.onnx_ml_pb2.ModelProto
            An ONNX model (type: ModelProto) which is equivalent to the input xgboost model.
        """
        auto_generated_initial_types = None
        if not initial_types:
            auto_generated_initial_types = self._generate_initial_types(X_sample)

        model_types = []
        if str(type(self.estimator)).startswith("<class 'xgboost.sklearn."):
            model_types.append(type(self.estimator))

        if model_types:
            if xgboost.sklearn.XGBClassifier in model_types:
                update_registered_converter(
                    xgboost.XGBClassifier,
                    "XGBoostXGBClassifier",
                    calculate_linear_classifier_output_shapes,
                    convert_xgboost,
                    options={"nocl": [True, False], "zipmap": [True, False]},
                )
            elif xgboost.sklearn.XGBRegressor in model_types:
                update_registered_converter(
                    xgboost.XGBRegressor,
                    "XGBoostXGBRegressor",
                    calculate_linear_regressor_output_shapes,
                    convert_xgboost,
                )
            if initial_types:
                return convert_sklearn(
                    self.estimator, initial_types=initial_types, **kwargs
                )
            else:
                try:
                    return convert_sklearn(
                        self.estimator,
                        initial_types=auto_generated_initial_types,
                        **kwargs,
                    )
                except:
                    raise ValueError(
                        "`initial_types` can not be autodetected. Please directly pass `initial_types`."
                    )
        else:
            # xgboost api
            if initial_types:
                return onnxmltools.convert_xgboost(
                    self.estimator,
                    initial_types=initial_types,
                    target_opset=kwargs.pop("target_opset", None),
                    targeted_onnx=onnx.__version__,
                    **kwargs,
                )
            else:
                try:
                    return onnxmltools.convert_xgboost(
                        self.estimator,
                        initial_types=auto_generated_initial_types,
                        target_opset=kwargs.pop("target_opset", None),
                        targeted_onnx=onnx.__version__,
                        **kwargs,
                    )
                except:
                    raise ValueError(
                        "`initial_types` can not be autodetected. Please directly pass `initial_types`."
                    )

    @runtime_dependency(
        module="skl2onnx.common.data_types",
        object="FloatTensorType",
        install_from=OptionalDependency.ONNX,
    )
    def _generate_initial_types(self, X_sample: Any) -> List:
        """Auto generate intial types.

        Parameters
        ----------
        X_sample: (Any)
            Train data.

        Returns
        -------
        List
            Initial types.
        """
        if hasattr(self.estimator, "n_features_in_"):
            # sklearn api
            n_cols = self.estimator.n_features_in_
            return [("input", FloatTensorType([None, n_cols]))]
        elif hasattr(self.estimator, "feature_names") and self.estimator.feature_names:
            # xgboost learning api
            n_cols = len(self.estimator.feature_names)
            return [("input", FloatTensorType([None, n_cols]))]
        if X_sample is None:
            raise ValueError(
                " At least one of `X_sample` or `initial_types` must be provided."
            )
        if (
            X_sample is not None
            and hasattr(X_sample, "shape")
            and len(X_sample.shape) >= 2
        ):
            auto_generated_initial_types = [
                ("input", FloatTensorType([None, X_sample.shape[1]]))
            ]
        else:
            raise ValueError(
                "`initial_types` can not be detected. Please directly pass initial_types."
            )
        return auto_generated_initial_types


class PytorchOnnxModelSerializer(OnnxModelSerializer):
    """Converts Pytorch model into onnx format."""

    def __init__(self):
        super().__init__()

    @runtime_dependency(module="torch", install_from=OptionalDependency.PYTORCH)
    def serialize(
        self,
        estimator,
        model_path: str,
        X_sample: Optional[
            Union[
                Dict,
                str,
                List,
                Tuple,
                np.ndarray,
                pd.core.series.Series,
                pd.core.frame.DataFrame,
            ]
        ] = None,
        **kwargs,
    ):
        """
        Exports the given Pytorch model into ONNX format.

        Parameters
        ----------
        path: str, default to None
            Path to save the serialized model.
        onnx_args: (tuple or torch.Tensor), default to None
            Contains model inputs such that model(onnx_args) is a valid
            invocation of the model. Can be structured either as: 1) ONLY A
            TUPLE OF ARGUMENTS; 2) A TENSOR; 3) A TUPLE OF ARGUMENTS ENDING
            WITH A DICTIONARY OF NAMED ARGUMENTS
        X_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]. Defaults to None.
            A sample of input data that will be used to generate input schema and detect onnx_args.
        kwargs:
            input_names: (List[str], optional). Defaults to ["input"].
                Names to assign to the input nodes of the graph, in order.
            output_names: (List[str], optional). Defaults to ["output"].
                Names to assign to the output nodes of the graph, in order.
            dynamic_axes: (dict, optional). Defaults to None.
                Specify axes of tensors as dynamic (i.e. known only at run-time).

        Returns
        -------
        None
            Nothing

        Raises
        ------
        AssertionError
            if onnx module is not support by the current version of torch
        ValueError
            if X_sample is not provided
            if path is not provided
        """
        onnx_args = kwargs.get("onnx_args", None)
        input_names = kwargs.get("input_names", ["input"])
        output_names = kwargs.get("output_names", ["output"])
        dynamic_axes = kwargs.get("dynamic_axes", None)

        assert hasattr(torch, "onnx"), (
            f"This version of pytorch {torch.__version__} does not appear to support onnx "
            "conversion."
        )

        if onnx_args is None:
            if X_sample is not None:
                logger.warning(
                    "Since `onnx_args` is not provided, `onnx_args` is "
                    "detected from `X_sample` to export pytorch model as onnx."
                )
                onnx_args = X_sample
            else:
                raise ValueError(
                    "`onnx_args` can not be detected. The parameter `onnx_args` must be provided to export pytorch model as onnx."
                )

        if not model_path:
            raise ValueError(
                "The parameter `model_path` must be provided to save the model file."
            )

        torch.onnx.export(
            estimator,
            args=onnx_args,
            f=model_path,
            input_names=input_names,
            output_names=output_names,
            dynamic_axes=dynamic_axes,
        )


class TensorFlowOnnxModelSerializer(OnnxModelSerializer):
    """Converts Tensorflow model into onnx format."""

    def __init__(self):
        super().__init__()

    @runtime_dependency(module="tf2onnx", install_from=OptionalDependency.ONNX)
    @runtime_dependency(
        module="tensorflow",
        short_name="tf",
        install_from=OptionalDependency.TENSORFLOW,
    )
    def serialize(
        self,
        estimator,
        model_path: str = None,
        X_sample: Optional[
            Union[
                Dict,
                str,
                List,
                Tuple,
                np.ndarray,
                pd.core.series.Series,
                pd.core.frame.DataFrame,
            ]
        ] = None,
        **kwargs,
    ):
        """
        Exports the given Tensorflow model into ONNX format.

        Parameters
        ----------
        model_path: str, default to None
            Path to save the serialized model.
        X_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]. Defaults to None.
            A sample of input data that will be used to generate input schema and detect input_signature.


        Returns
        -------
        None
            Nothing

        Raises
        ------
        ValueError
            if model_path is not provided
        """
        opset_version = kwargs.get("opset_version", None)
        input_signature = kwargs.get("input_signature", None)

        if not model_path:
            raise ValueError(
                "The parameter `model_path` must be provided to save the model file."
            )
        if input_signature is None:
            if hasattr(estimator, "input_shape"):
                if not isinstance(estimator.input, list):
                    # single input
                    detected_input_signature = (
                        tf.TensorSpec(
                            estimator.input_shape,
                            dtype=estimator.input.dtype,
                            name="input",
                        ),
                    )
                else:
                    # multiple input
                    detected_input_signature = []
                    for i in range(len(estimator.input)):
                        detected_input_signature.append(
                            tf.TensorSpec(
                                estimator.input_shape[i],
                                dtype=estimator.input[i].dtype,
                            )
                        )

            elif X_sample is not None and hasattr(X_sample, "shape"):
                logger.warning(
                    "Since `input_signature` is not provided, `input_signature` is "
                    "detected from `X_sample` to export tensorflow model as "
                    "onnx."
                )
                X_sample_shape = list(X_sample.shape)
                X_sample_shape[0] = None
                detected_input_signature = (
                    tf.TensorSpec(X_sample_shape, dtype=X_sample.dtype, name="input"),
                )
            else:
                raise ValueError(
                    "The parameter `input_signature` must be provided to export "
                    "tensorflow model as onnx."
                )
            try:
                tf2onnx.convert.from_keras(
                    estimator,
                    input_signature=detected_input_signature,
                    opset=opset_version,
                    output_path=model_path,
                )
            except:
                raise ValueError(
                    "`input_signature` can not be autodetected. The parameter `input_signature` must be provided to export "
                    "tensorflow model as onnx."
                )

        else:
            tf2onnx.convert.from_keras(
                estimator,
                input_signature=input_signature,
                opset=opset_version,
                output_path=model_path,
            )


class OnnxModelSaveSERDE(OnnxModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_ONNX


class CloudpickleModelSaveSERDE(CloudPickleModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_CLOUDPICKLE


class JoblibModelSaveSERDE(JobLibModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_JOBLIB


class SparkModelSaveSERDE(SparkModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_SPARK


class HuggingFacePipelineSaveSERDE(HuggingFaceModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_HUGGINGFACE


class TorchScriptModelSaveSERDE(TorchScriptModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_TORHCSCRIPT


class PyTorchModelSaveSERDE(PyTorchModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_TORCH


class PyTorchOnnxModelSaveSERDE(PytorchOnnxModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_TORCH_ONNX


class TensorFlowModelSaveSERDE(TensorFlowModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_TF


class TensorFlowOnnxModelSaveSERDE(TensorFlowOnnxModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_TF_ONNX


class SklearnOnnxModelSaveSERDE(SklearnOnnxModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_SKLEARN_ONNX


class LightGBMModelSaveSERDE(LightGBMModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_LIGHTGBM


class LightGBMOnnxModelSaveSERDE(LightGBMOnnxModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_LIGHTGBM_ONNX


class XgboostJsonModelSaveSERDE(XgboostJsonModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_XGBOOST


class XgboostUbjModelSaveSERDE(XgboostUbjModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_XGBOOST_UBJ


class XgboostTxtModelSaveSERDE(XgboostTxtModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_XGBOOST_TXT


class XgboostOnnxModelSaveSERDE(XgboostOnnxModelSerializer, ModelDeserializer):
    name = MODEL_SERIALIZATION_TYPE_XGBOOST_ONNX


class ModelSerializerFactory:
    """Model Serializer Factory.

    Returns
    -------
    model_save_serde: Intance of `ads.model.SERDE`".
    """

    _factory = {}
    _factory[MODEL_SERIALIZATION_TYPE_CLOUDPICKLE] = CloudpickleModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_ONNX] = OnnxModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_TORHCSCRIPT] = TorchScriptModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_TORCH] = PyTorchModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_TORCH_ONNX] = PyTorchOnnxModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_TF] = TensorFlowModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_TF_ONNX] = TensorFlowOnnxModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_JOBLIB] = JoblibModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_SKLEARN_ONNX] = SklearnOnnxModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_LIGHTGBM] = LightGBMModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_LIGHTGBM_ONNX] = LightGBMOnnxModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_XGBOOST] = XgboostJsonModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_XGBOOST_UBJ] = XgboostUbjModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_XGBOOST_TXT] = XgboostTxtModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_XGBOOST_ONNX] = XgboostOnnxModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_SPARK] = SparkModelSaveSERDE
    _factory[MODEL_SERIALIZATION_TYPE_HUGGINGFACE] = HuggingFacePipelineSaveSERDE

    @classmethod
    def get(cls, se: str):
        serde = cls._factory.get(se, None)
        if serde:
            return serde()
        else:
            raise ValueError(
                f"This {se} format is not supported."
                f"Currently support the following format: {SUPPORTED_MODEL_SERIALIZERS}."
            )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from enum import Enum, auto
from typing import Any, List, Dict

from fsspec.core import OpenFile


class Options(Enum):
    FILE_NAME = auto()
    FILE_METADATA = auto()


class OptionHandler:
    def __init__(self, dataloader: "ads.text_dataset.dataset.DataLoader") -> None:
        self.dataloader = dataloader

    def handle(self, fhandler: OpenFile, spec: Any) -> Any:
        raise NotImplementedError()


class FileOption(OptionHandler):
    def handle(self, fhandler: OpenFile, spec: Any) -> Any:
        return fhandler.path


class MetadataOption(OptionHandler):
    def handle(self, fhandler: OpenFile, spec: Dict) -> List:
        metadata = self.dataloader.processor.get_metadata(fhandler)
        return [metadata.get(k, None) for k in spec["extract"]]


class OptionFactory:

    option_handlers = {
        Options.FILE_NAME: FileOption,
        Options.FILE_METADATA: MetadataOption,
    }

    @staticmethod
    def option_handler(option: Options) -> OptionHandler:
        handler = OptionFactory.option_handlers.get(option, None)
        if handler is None:
            raise RuntimeError(f"Option {option} Not Recognized.")
        return handler

    @classmethod
    def register_option(cls, option: Options, handler) -> None:
        cls.option_handlers[option] = handler
